@@ -98,7 +98,7 @@ void set_absolute(InputInfoPtr pInfo, Bool absolute)
 static void wcmSendButtons(InputInfoPtr pInfo, int buttons,
 			   int first_val, int num_vals, int *valuators)
 {
-	int button, mask, first_button;
+	unsigned int button, mask, first_button;
 	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
 	WacomCommonPtr common = priv->common;
 	DBG(6, priv, "buttons=%d\n", buttons);
@@ -134,7 +134,7 @@ static void wcmSendButtons(InputInfoPtr pInfo, int buttons,
 
 	for (button = first_button; button < WCM_MAX_BUTTONS; button++)
 	{
-		mask = 1 << button;
+		mask = 1u << button;
 		if ((mask & priv->oldState.buttons) != (mask & buttons))
 			sendAButton(pInfo, button, (mask & buttons),
 					first_val, num_vals, valuators);
@@ -361,6 +361,7 @@ static void sendWheelStripEvents(InputInfoPtr pInfo, const WacomDeviceState* ds,
 				 int first_val, int num_vals, int *valuators)
 {
 	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
+	WacomCommonPtr common = priv->common;
 	int delta = 0, idx = 0;
 
 	DBG(10, priv, "\n");
@@ -396,7 +397,7 @@ static void sendWheelStripEvents(InputInfoPtr pInfo, const WacomDeviceState* ds,
 	}
 
 	/* emulate events for left touch ring */
-	delta = getScrollDelta(ds->abswheel, priv->oldState.abswheel, MAX_PAD_RING, AXIS_INVERT);
+	delta = getScrollDelta(ds->abswheel, priv->oldState.abswheel, common->wcmMaxRing, AXIS_INVERT);
 	idx = getWheelButton(delta, WHEEL_ABS_UP, WHEEL_ABS_DN);
 	if (idx >= 0 && IsPad(priv) && priv->oldState.proximity == ds->proximity)
 	{
@@ -406,7 +407,7 @@ static void sendWheelStripEvents(InputInfoPtr pInfo, const WacomDeviceState* ds,
 	}
 
 	/* emulate events for right touch ring */
-	delta = getScrollDelta(ds->abswheel2, priv->oldState.abswheel2, MAX_PAD_RING, AXIS_INVERT);
+	delta = getScrollDelta(ds->abswheel2, priv->oldState.abswheel2, common->wcmMaxRing, AXIS_INVERT);
 	idx = getWheelButton(delta, WHEEL2_ABS_UP, WHEEL2_ABS_DN);
 	if (idx >= 0 && IsPad(priv) && priv->oldState.proximity == ds->proximity)
 	{
@@ -706,7 +707,8 @@ void wcmSendEvents(InputInfoPtr pInfo, const WacomDeviceState* ds)
 	valuators[3] = v3;
 	valuators[4] = v4;
 	valuators[5] = v5;
-	valuators[6] = v6;
+	if (priv->naxes > 6)
+		valuators[6] = v6;
 
 	if (type == PAD_ID)
 		wcmSendPadEvents(pInfo, ds, 3, priv->naxes - 3, &valuators[3]); /* pad doesn't post x/y/z */
@@ -852,29 +854,38 @@ static Bool check_arbitrated_control(InputInfoPtr pInfo, WacomDeviceStatePtr ds)
 
 	if (IsPad(priv)) {
 		/* Pad may never be the "active" pointer controller */
+		DBG(6, priv, "Event from pad; not yielding pointer control\n.");
 		return FALSE;
 	}
 
 	if (active == NULL || active->oldState.device_id == ds->device_id) {
-		DBG(11, priv, "Same device ID as active; allowing access.\n");
+		DBG(11, priv, "Event from active device; maintaining pointer control.\n");
 		return TRUE;
 	}
-	else if (IsCursor(active) && IsTouch(priv)) {
-		/* Cursor devices are often left idle in range, so allow touch to
-		 * grab control if the tool has not been used for some time.
+	else if (IsCursor(active)) {
+		/* Cursor devices are often left idle in range, so allow other devices
+		 * to grab control if the tool has not been used for some time.
 		 */
-		return (ds->time - active->oldState.time > 100);
+		Bool yield = (ds->time - active->oldState.time > 100) && (active->oldState.buttons == 0);
+		DBG(6, priv, "Currently-active cursor %s idle; %s pointer control.\n",
+		    yield ? "is" : "is not", yield ? "yielding" : "not yielding");
+		return yield;
 	}
-	else if (IsTouch(active) && IsCursor(priv)) {
+	else if (IsCursor(priv)) {
 		/* An otherwise idle cursor may still occasionally jitter and send
-		 * events while the user is making active touches. Do not allow
-		 * the cursor to grab control in this particular case.
+		 * events while the user is actively using other tools or touching
+		 * the device. Do not allow the cursor to grab control in this
+		 * particular case.
 		 */
+		DBG(6, priv, "Event from non-active cursor; not yielding pointer control.\n");
 		return FALSE;
 	}
 	else {
 		/* Non-touch input has priority over touch in general */
-		return !IsTouch(priv);
+		Bool yield = !IsTouch(priv);
+		DBG(6, priv, "Event from non-active %s device; %s pointer control.\n",
+		    yield ? "non-touch" : "touch", yield ? "yielding" : "not yielding");
+		return yield;
 	}
 }
 
@@ -1045,7 +1056,7 @@ rebasePressure(const WacomDevicePtr priv, const WacomDeviceState *ds)
 
 /**
  * Instead of reporting the raw pressure, we normalize
- * the pressure from 0 to FILTER_PRESSURE_RES. This is
+ * the pressure from 0 to maxCurve. This is
  * mainly to deal with the case where heavily used
  * stylus may have a "pre-loaded" initial pressure. To
  * do so, we keep the in-prox pressure and subtract it
@@ -1070,14 +1081,14 @@ normalizePressure(const WacomDevicePtr priv, const int raw_pressure)
 		p -= priv->minPressure;
 		range_left -= priv->minPressure;
 	}
-	/* normalize pressure to 0..FILTER_PRESSURE_RES */
+	/* normalize pressure to 0..maxCurve */
 	if (range_left >= 1)
 		pressure = xf86ScaleAxis(p,
-					 FILTER_PRESSURE_RES, 0,
+					 priv->maxCurve, 0,
 					 range_left,
 					 0);
 	else
-		pressure = FILTER_PRESSURE_RES;
+		pressure = priv->maxCurve;
 
 	return (int)pressure;
 }
@@ -1106,8 +1117,8 @@ setPressureButton(const WacomDevicePtr priv, int buttons, const int pressure)
 		{
 			/* don't set it off if it is within the tolerance
 			   and threshold is larger than the tolerance */
-			if ((common->wcmThreshold > THRESHOLD_TOLERANCE) &&
-			    (pressure > common->wcmThreshold - THRESHOLD_TOLERANCE))
+			if ((common->wcmThreshold > (priv->maxCurve * THRESHOLD_TOLERANCE)) &&
+			    (pressure > common->wcmThreshold - (priv->maxCurve * THRESHOLD_TOLERANCE)))
 				buttons |= button;
 		}
 	}
@@ -1185,7 +1196,7 @@ static void commonDispatchDevice(InputInfoPtr pInfo,
 	if (priv->serial && filtered.serial_num != priv->serial)
 	{
 		DBG(10, priv, "serial number"
-			" is %u but your system configured %u",
+			" is %u but your system configured %u\n",
 			filtered.serial_num, priv->serial);
 		return;
 	}
@@ -1239,15 +1250,6 @@ static void commonDispatchDevice(InputInfoPtr pInfo,
 		}
 		filtered.pressure = applyPressureCurve(priv,&filtered);
 	}
-	else if (IsCursor(priv) && !priv->oldCursorHwProx)
-	{
-		/* initial current max distance for Intuos series */
-		if ((TabletHasFeature(common, WCM_ROTATION)) ||
-				(TabletHasFeature(common, WCM_DUALINPUT)))
-			common->wcmMaxCursorDist = common->wcmMaxDist;
-		else
-			common->wcmMaxCursorDist = 0;
-	}
 
 	/* Store cursor hardware prox for next use */
 	if (IsCursor(priv))
@@ -1266,7 +1268,7 @@ static void commonDispatchDevice(InputInfoPtr pInfo,
 		double delty = filtered.y - priv->oldState.y;
 
 		/* less than one device coordinate movement? */
-		if (abs(deltx)<1 && abs(delty)<1)
+		if (fabs(deltx)<1 && fabs(delty)<1)
 		{
 			/* We have no other data in this event, skip */
 			if (suppress == SUPPRESS_NON_MOTION)
@@ -1284,20 +1286,15 @@ static void commonDispatchDevice(InputInfoPtr pInfo,
 	/* force out-prox when distance is outside wcmCursorProxoutDist for pucks */
 	if (IsCursor(priv))
 	{
-		if (common->wcmProtocolLevel == WCM_PROTOCOL_5)
-		{
-			/* protocol 5 distance starts from the MaxDist
-			 * when getting in the prox.
-			 */
-			if (common->wcmMaxCursorDist > filtered.distance)
-				common->wcmMaxCursorDist = filtered.distance;
-		}
-		else
-		{
-			/* protocol 4 distance is 0 when getting in the prox */
-			if (common->wcmMaxCursorDist < filtered.distance)
-				common->wcmMaxCursorDist = filtered.distance;
-		}
+		/* Assume the the user clicks the puck buttons while
+		 * it is resting on the tablet. This works for both
+		 * tablets that have a normal distance scale (protocol
+		 * 5) as well as those with an inverted scale (protocol
+		 * 4 for many many kernel versions).
+		 */
+		if (filtered.buttons)
+			common->wcmMaxCursorDist = filtered.distance;
+
 		DBG(10, common, "Distance over"
 				" the tablet: %d, ProxoutDist: %d current"
 				" min/max %d hard prox: %d\n",
@@ -1306,21 +1303,23 @@ static void commonDispatchDevice(InputInfoPtr pInfo,
 				common->wcmMaxCursorDist,
 				ds->proximity);
 
-		if (priv->oldState.proximity)
-		{
-			if (abs(filtered.distance - common->wcmMaxCursorDist)
-					> common->wcmCursorProxoutDist)
-				filtered.proximity = 0;
-		}
-		/* once it is out. Don't let it in until a hard in */
-		/* or it gets inside wcmCursorProxoutDist */
-		else
-		{
-			if (abs(filtered.distance - common->wcmMaxCursorDist) >
-					common->wcmCursorProxoutDist && ds->proximity)
-				return;
-			if (!ds->proximity)
-				return;
+		if (common->wcmMaxCursorDist) {
+			if (priv->oldState.proximity)
+			{
+				if (abs(filtered.distance - common->wcmMaxCursorDist)
+						> common->wcmCursorProxoutDist)
+					filtered.proximity = 0;
+			}
+			/* once it is out. Don't let it in until a hard in */
+			/* or it gets inside wcmCursorProxoutDist */
+			else
+			{
+				if (abs(filtered.distance - common->wcmMaxCursorDist) >
+						common->wcmCursorProxoutDist && ds->proximity)
+					return;
+				if (!ds->proximity)
+					return;
+			}
 		}
 	}
 	wcmSendEvents(pInfo, &filtered);
@@ -1351,7 +1350,7 @@ int wcmInitTablet(InputInfoPtr pInfo, const char* id, float version)
 	if (common->wcmThreshold <= 0 && IsPen(priv))
 	{
 		/* Threshold for counting pressure as a button */
-		common->wcmThreshold = DEFAULT_THRESHOLD;
+		common->wcmThreshold = priv->maxCurve * DEFAULT_THRESHOLD;
 
 		xf86Msg(X_PROBED, "%s: using pressure threshold of %d for button 1\n",
 			pInfo->name, common->wcmThreshold);
@@ -1402,10 +1401,13 @@ static int applyPressureCurve(WacomDevicePtr pDev, const WacomDeviceStatePtr pSt
 	/* clip the pressure */
 	int p = max(0, pState->pressure);
 
-	p = min(FILTER_PRESSURE_RES, p);
+	p = min(pDev->maxCurve, p);
 
 	/* apply pressure curve function */
-	return pDev->pPressCurve[p];
+	if (pDev->pPressCurve == NULL)
+		return p;
+	else
+		return pDev->pPressCurve[p];
 }
 
 /*****************************************************************************
@@ -1475,6 +1477,9 @@ void wcmFreeCommon(WacomCommonPtr *ptr)
 {
 	WacomCommonPtr common = *ptr;
 
+	if (!common)
+		return;
+
 	DBG(10, common, "common refcount dec to %d\n", common->refcnt - 1);
 	if (--common->refcnt == 0)
 	{
