@@ -1,6 +1,6 @@
 /*
  * Copyright 1995-2002 by Frederic Lepied, France. <Lepied@XFree86.org>
- * Copyright 2002-2010 by Ping Cheng, Wacom. <pingc@wacom.com>
+ * Copyright 2002-2013 by Ping Cheng, Wacom. <pingc@wacom.com>
  *                                                                            
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -63,6 +63,7 @@ static int wcmAllocate(InputInfoPtr pInfo)
 	priv->pInfo = pInfo;
 	priv->common = common;       /* common info pointer */
 	priv->oldCursorHwProx = 0;   /* previous cursor hardware proximity */
+	priv->maxCurve = FILTER_PRESSURE_RES;
 	priv->nPressCtrl [0] = 0;    /* pressure curve x0 */
 	priv->nPressCtrl [1] = 0;    /* pressure curve y0 */
 	priv->nPressCtrl [2] = 100;  /* pressure curve x1 */
@@ -71,26 +72,23 @@ static int wcmAllocate(InputInfoPtr pInfo)
 	/* Default button and expresskey values, offset buttons 4 and higher
 	 * by the 4 scroll buttons. */
 	for (i=0; i<WCM_MAX_BUTTONS; i++)
-		priv->button[i] = (i < 3) ? i + 1 : i + 5;
+		priv->button_default[i] = (i < 3) ? i + 1 : i + 5;
 
-	priv->nbuttons = WCM_MAX_BUTTONS;		/* Default number of buttons */
-	priv->relup = 5;			/* Default relative wheel up event */
-	priv->reldn = 4;			/* Default relative wheel down event */
+	priv->nbuttons = WCM_MAX_BUTTONS;       /* Default number of buttons */
+	priv->wheel_default[WHEEL_REL_UP] = 5;
+	priv->wheel_default[WHEEL_REL_DN] = 4;
 	/* wheel events are set to 0, but the pad overwrites this default
 	 * later in wcmParseOptions, when we have IsPad() available */
-	priv->wheelup = 0;			/* Default absolute wheel up event */
-	priv->wheeldn = 0;			/* Default absolute wheel down event */
-	priv->wheel2up = 0;                     /* Default absolute wheel2 up event */
-	priv->wheel2dn = 0;                     /* Default absolute wheel2 down event */
-	priv->striplup = 4;			/* Default left strip up event */
-	priv->stripldn = 5;			/* Default left strip down event */
-	priv->striprup = 4;			/* Default right strip up event */
-	priv->striprdn = 5;			/* Default right strip down event */
+	priv->wheel_default[WHEEL_ABS_UP] = 0;
+	priv->wheel_default[WHEEL_ABS_DN] = 0;
+	priv->wheel_default[WHEEL2_ABS_UP] = 0;
+	priv->wheel_default[WHEEL2_ABS_DN] = 0;
+	priv->strip_default[STRIP_LEFT_UP] = 4;
+	priv->strip_default[STRIP_LEFT_DN] = 5;
+	priv->strip_default[STRIP_RIGHT_UP] = 4;
+	priv->strip_default[STRIP_RIGHT_DN] = 5;
 	priv->naxes = 6;			/* Default number of axes */
 
-	/* JEJ - throttle sampling code */
-	priv->throttleLimit = -1;
-
 	common->wcmDevices = priv;
 
 	/* tool */
@@ -102,6 +100,8 @@ static int wcmAllocate(InputInfoPtr pInfo)
 
 	/* timers */
 	priv->serial_timer = TimerSet(NULL, 0, 0, NULL, NULL);
+	priv->tap_timer = TimerSet(NULL, 0, 0, NULL, NULL);
+	priv->touch_timer = TimerSet(NULL, 0, 0, NULL, NULL);
 
 	return 1;
 
@@ -125,6 +125,8 @@ static void wcmFree(InputInfoPtr pInfo)
 		return;
 
 	TimerFree(priv->serial_timer);
+	TimerFree(priv->tap_timer);
+	TimerFree(priv->touch_timer);
 	free(priv->tool);
 	wcmFreeCommon(&priv->common);
 	free(priv);
@@ -132,7 +134,8 @@ static void wcmFree(InputInfoPtr pInfo)
 	pInfo->private = NULL;
 }
 
-static int wcmSetType(InputInfoPtr pInfo, const char *type)
+TEST_NON_STATIC int
+wcmSetType(InputInfoPtr pInfo, const char *type)
 {
 	WacomDevicePtr priv = pInfo->private;
 
@@ -168,7 +171,7 @@ static int wcmSetType(InputInfoPtr pInfo, const char *type)
 		goto invalid;
 
 	/* Set the device id of the "last seen" device on this tool */
-	priv->old_device_id = wcmGetPhyDeviceID(priv);
+	priv->oldState.device_id = wcmGetPhyDeviceID(priv);
 
 	if (!priv->tool)
 		return 0;
@@ -207,7 +210,7 @@ int wcmGetPhyDeviceID(WacomDevicePtr priv)
  * starts making sense again.
  */
 
-static char *default_options[] =
+static const char *default_options[] =
 {
 	"StopBits",    "1",
 	"DataBits",    "8",
@@ -234,6 +237,9 @@ static void wcmUninit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 
 	DBG(1, priv, "\n");
 
+	if (WACOM_DRIVER.active == priv)
+		WACOM_DRIVER.active = NULL;
+
 	/* Server 1.10 will UnInit all devices for us */
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
 	if (priv->isParent)
@@ -296,6 +302,94 @@ out:
 	xf86DeleteInput(pInfo, 0);
 }
 
+/**
+ * Splits a wacom device name into its constituent pieces. For instance,
+ * "Wacom Intuos Pro Finger touch" would be split into "Wacom Intuos Pro"
+ * (the base kernel device name), "Finger" (an descriptor of the specific
+ * event interface), and "touch" (a suffix added by this driver to indicate
+ * the specific tool).
+ */
+static void wcmSplitName(char* devicename, char *basename, char *subdevice, char *tool, size_t len)
+{
+	char *name = strdupa(devicename);
+	char *a, *b;
+
+	*basename = *subdevice = *tool = '\0';
+
+	a = strrchr(name, ' ');
+	if (a)
+	{
+		*a = '\0';
+		b = strrchr(name, ' ');
+
+		while (b)
+		{
+			if (!strcmp(b, " Pen") || !strcmp(b, " Finger") ||
+			    !strcmp(b, " Pad") || !strcmp(b, " Touch"))
+			{
+				*b = '\0';
+				strncpy(subdevice, b+1, len-1);
+				subdevice[len-1] = '\0';
+				b = strrchr(name, ' ');
+			}
+			else
+				b = NULL;
+		}
+		strncat(tool, a+1, len-1);
+	}
+	strncat(basename, name, len-1);
+}
+
+/**
+ * Determines if two input devices represent independent parts (stylus,
+ * eraser, pad) of the same underlying device. If the 'logical_only'
+ * flag is set, the function will only return true if the two devices
+ * are represented by the same logical device (i.e., share the same
+ * input device node). Otherwise, the function will attempt to determine
+ * if the two devices are part of the same physical tablet, such as
+ * when a tablet reports 'pen' and 'touch' through separate device
+ * nodes.
+ */
+static Bool wcmIsSiblingDevice(InputInfoPtr a, InputInfoPtr b, Bool logical_only)
+{
+	WacomDevicePtr privA = (WacomDevicePtr)a->private;
+	WacomDevicePtr privB = (WacomDevicePtr)b->private;
+
+	if (strcmp(a->drv->driverName, "wacom") || strcmp(b->drv->driverName, "wacom"))
+		return FALSE;
+
+	if (privA == privB)
+		return FALSE;
+
+	if (DEVICE_ID(privA->flags) == DEVICE_ID(privB->flags))
+		return FALSE;
+
+	if (!strcmp(privA->common->device_path, privB->common->device_path))
+		return TRUE;
+
+	if (!logical_only)
+	{
+		// TODO: Udev might provide more accurate data, but this should
+		// be good enough in practice.
+		const int len = 50;
+		char baseA[len], subA[len], toolA[len];
+		char baseB[len], subB[len], toolB[len];
+		wcmSplitName(privA->name, baseA, subA, toolA, len);
+		wcmSplitName(privB->name, baseB, subB, toolB, len);
+
+		if (strcmp(baseA, baseB))
+		{
+			// Fallback for (arbitrary) static xorg.conf device names
+			return (privA->common->tablet_id == privB->common->tablet_id);
+		}
+
+		if (strlen(subA) != 0 && strlen(subB) != 0)
+			return TRUE;
+	}
+
+	return FALSE;
+}
+
 /* wcmMatchDevice - locate matching device and merge common structure. If an
  * already initialized device shares the same device file and driver, remove
  * the new device's "common" struct and point to the one of the already
@@ -320,9 +414,7 @@ static Bool wcmMatchDevice(InputInfoPtr pLocal, WacomCommonPtr *common_return)
 	{
 		WacomDevicePtr privMatch = (WacomDevicePtr)pMatch->private;
 
-		if ((pLocal != pMatch) &&
-				strstr(pMatch->drv->driverName, "wacom") &&
-				!strcmp(privMatch->common->device_path, common->device_path))
+		if (wcmIsSiblingDevice(pLocal, pMatch, TRUE))
 		{
 			DBG(2, priv, "port share between %s and %s\n",
 					pLocal->name, pMatch->name);
@@ -376,7 +468,7 @@ wcmInitModel(InputInfoPtr pInfo)
 	float version;
 
 	/* Initialize the tablet */
-	if(common->wcmDevCls->Init(pInfo, id, &version) != Success ||
+	if(common->wcmDevCls->Init(pInfo, id, ARRAY_SIZE(id), &version) != Success ||
 		wcmInitTablet(pInfo, id, version) != Success)
 		return FALSE;
 
@@ -384,48 +476,70 @@ wcmInitModel(InputInfoPtr pInfo)
 }
 
 /**
- * Link the touch tool to the pen of the same device
- * so we can arbitrate the events when posting them.
+ * Lookup to find the associated pen and touch for the same device.
+ * Store touch tool in wcmTouchDevice for pen and touch, respectively,
+ * of the same device. Update TabletFeature to indicate it is a hybrid
+ * of touch and pen.
+ *
+ * @return True if found a touch tool for hybrid devices.
+ * false otherwise.
  */
-static void wcmLinkTouchAndPen(InputInfoPtr pInfo)
+static Bool wcmLinkTouchAndPen(InputInfoPtr pInfo)
 {
 	WacomDevicePtr priv = pInfo->private;
 	WacomCommonPtr common = priv->common;
 	InputInfoPtr device = xf86FirstLocalDevice();
 	WacomCommonPtr tmpcommon = NULL;
 	WacomDevicePtr tmppriv = NULL;
-	Bool touch_device_assigned = FALSE;
 
 	/* Lookup to find the associated pen and touch */
 	for (; device != NULL; device = device->next)
 	{
-		if (!strcmp(device->drv->driverName, "wacom"))
+		if (!wcmIsSiblingDevice(pInfo, device, FALSE))
+			continue;
+
+		tmppriv = (WacomDevicePtr) device->private;
+		tmpcommon = tmppriv->common;
+
+		DBG(4, priv, "Considering link with %s...\n", tmppriv->name);
+
+		/* already linked devices */
+		if (tmpcommon->wcmTouchDevice && IsTablet(tmppriv))
 		{
-			tmppriv = (WacomDevicePtr) device->private;
-			tmpcommon = tmppriv->common;
-			touch_device_assigned = (common->wcmTouchDevice ||
-						tmpcommon->wcmTouchDevice);
+			DBG(4, priv, "A link is already in place. Ignoring.\n");
+			continue;
+		}
 
-			/* skip the same tool or already linked devices */
-			if ((tmppriv == priv) || touch_device_assigned)
-				continue;
+		if (IsTouch(tmppriv))
+		{
+			common->wcmTouchDevice = tmppriv;
+			tmpcommon->wcmTouchDevice = tmppriv;
+		}
+		else if (IsTouch(priv))
+		{
+			common->wcmTouchDevice = priv;
+			tmpcommon->wcmTouchDevice = priv;
+		}
+		else
+		{
+			DBG(4, priv, "A link is not necessary. Ignoring.\n");
+		}
 
-			if (tmpcommon->tablet_id == common->tablet_id)
-			{
-				if (IsTouch(tmppriv) && IsPen(priv))
-					common->wcmTouchDevice = tmppriv;
-				else if (IsTouch(priv) && IsPen(tmppriv))
-					tmpcommon->wcmTouchDevice = priv;
+		if ((common->wcmTouchDevice && IsTablet(priv)) ||
+			(tmpcommon->wcmTouchDevice && IsTablet(tmppriv)))
+		{
+			TabletSetFeature(common, WCM_PENTOUCH);
+			TabletSetFeature(tmpcommon, WCM_PENTOUCH);
+		}
 
-				if (common->wcmTouchDevice ||
-						tmpcommon->wcmTouchDevice)
-				{
-					TabletSetFeature(common, WCM_PENTOUCH);
-					TabletSetFeature(tmpcommon, WCM_PENTOUCH);
-				}
-			}
+		if (common->wcmTouchDevice)
+		{
+			DBG(4, priv, "Link created!\n");
+			return TRUE;
 		}
 	}
+	DBG(4, priv, "No suitable device to link with found.\n");
+	return FALSE;
 }
 
 /**
@@ -440,7 +554,9 @@ static void wcmLinkTouchAndPen(InputInfoPtr pInfo)
 static int wcmIsHotpluggedDevice(InputInfoPtr pInfo)
 {
 	char *source = xf86CheckStrOption(pInfo->options, "_source", "");
-	return !strcmp(source, "_driver/wacom");
+	int matches = (strcmp(source, "_driver/wacom") == 0);
+	free(source);
+	return matches;
 }
 
 /* wcmPreInit - called for each input devices with the driver set to
@@ -480,8 +596,8 @@ static int wcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 {
 	WacomDevicePtr priv = NULL;
 	WacomCommonPtr common = NULL;
-	const char*	type;
-	char*		device, *oldname;
+	char		*type, *device;
+	char		*oldname = NULL;
 	int		need_hotplug = 0, is_dependent = 0;
 
 	gWacomModule.wcmDrv = drv;
@@ -531,7 +647,7 @@ static int wcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 
 	common->debugLevel = xf86SetIntOption(pInfo->options,
 					      "CommonDBG", common->debugLevel);
-	oldname = pInfo->name;
+	oldname = strdup(pInfo->name);
 
 	if (wcmIsHotpluggedDevice(pInfo))
 		is_dependent = 1;
@@ -542,6 +658,7 @@ static int wcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 		char *new_name;
 		if (asprintf(&new_name, "%s %s", pInfo->name, type) == -1)
 			new_name = strdup(pInfo->name);
+		free(pInfo->name);
 		pInfo->name = priv->name = new_name;
 	}
 
@@ -567,18 +684,17 @@ static int wcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 		wcmHotplugOthers(pInfo, oldname);
 	}
 
-	if (pInfo->fd != -1)
-	{
-		close(pInfo->fd);
-		pInfo->fd = -1;
-	}
+	wcmClose(pInfo);
 
-	/* only link them once per port. We need to try for both pen and touch
-	 * since we do not know which tool (touch or pen) will be added first.
+	/* only link them once per port. We need to try for both tablet tool
+	 * and touch since we do not know which tool will be added first.
 	 */
-	if (IsTouch(priv) || (IsPen(priv) && !common->wcmTouchDevice))
+	if (IsTouch(priv) || (IsTablet(priv) && !common->wcmTouchDevice))
 		wcmLinkTouchAndPen(pInfo);
 
+	free(type);
+	free(oldname);
+
 	return Success;
 
 SetupProc_fail:
@@ -586,16 +702,13 @@ SetupProc_fail:
 	if (common && priv)
 		common->wcmDevices = priv->next;
 
-	if (pInfo->fd != -1)
-	{
-		close(pInfo->fd);
-		pInfo->fd = -1;
-	}
-
+	wcmClose(pInfo);
+	free(type);
+	free(oldname);
 	return BadMatch;
 }
 
-InputDriverRec WACOM =
+static InputDriverRec WACOM =
 {
 	1,             /* driver version */
 	"wacom",       /* driver name */
@@ -604,7 +717,10 @@ InputDriverRec WACOM =
 	wcmUninit, /* un-init */
 	NULL,          /* module */
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
-	default_options
+	default_options,
+#endif
+#ifdef XI86_DRV_CAP_SERVER_FD
+	XI86_DRV_CAP_SERVER_FD,
 #endif
 };
 
@@ -621,6 +737,9 @@ static pointer wcmPlug(pointer module, pointer options, int* errmaj,
 		int* errmin)
 {
 	xf86AddInputDriver(&WACOM, module, 0);
+
+	usbListModels();
+
 	return module;
 }
 
