@@ -287,6 +287,14 @@ static struct WacomModelDesc
 	{ WACOM_VENDOR_ID, 0x57, 200000, 200000, &usbCintiqV5,   "DTK2241"		},
 	{ WACOM_VENDOR_ID, 0x59, 200000, 200000, &usbCintiqV5,   "DTH2242"		},
 	{ WACOM_VENDOR_ID, 0x5B, 200000, 200000, &usbCintiqV5,   "Cintiq 22HDT"		},
+	{ WACOM_VENDOR_ID, 0x34A, 200000, 200000, &usbCintiqV5,  "Mobilestudio Pro 13"	}, /* Touch */
+	{ WACOM_VENDOR_ID, 0x34B, 200000, 200000, &usbCintiqV5,  "MobileStudio Pro 16"	}, /* Touch */
+	{ WACOM_VENDOR_ID, 0x34D, 200000, 200000, &usbCintiqV5,  "MobileStudio Pro 13"	},
+	{ WACOM_VENDOR_ID, 0x34E, 200000, 200000, &usbCintiqV5,  "MobileStudio Pro 16"	},
+	{ WACOM_VENDOR_ID, 0x34F, 200000, 200000, &usbCintiqV5,  "Cintiq 13 FHD Pro"	},
+	{ WACOM_VENDOR_ID, 0x350, 200000, 200000, &usbCintiqV5,  "Cintiq 16 UHD Pro"	},
+	{ WACOM_VENDOR_ID, 0x353, 200000, 200000, &usbCintiqV5,  "Cintiq 13 FHD Pro"	}, /* Touch */
+	{ WACOM_VENDOR_ID, 0x354, 200000, 200000, &usbCintiqV5,  "Cintiq 16 UHD Pro"	}, /* Touch */
 
 	{ WACOM_VENDOR_ID, 0x90, 100000, 100000, &usbTabletPC,   "TabletPC 0x90"	},
 	{ WACOM_VENDOR_ID, 0x93, 100000, 100000, &usbTabletPC,   "TabletPC 0x93"	},
@@ -401,6 +409,14 @@ static Bool usbWcmInit(InputInfoPtr pInfo, char* id, size_t id_len, float *versi
 
 	DBG(1, priv, "initializing USB tablet\n");
 
+	/* fetch vendor, product, and model name */
+	if (ioctl(pInfo->fd, EVIOCGID, &sID) == -1 ||
+	    ioctl(pInfo->fd, EVIOCGNAME(id_len), id) == -1) {
+		xf86Msg(X_ERROR, "%s: failed to ioctl ID or name.\n",
+					pInfo->name);
+		return !Success;
+	}
+
 	if (!common->private &&
 	    !(common->private = calloc(1, sizeof(wcmUSBData))))
 	{
@@ -412,10 +428,6 @@ static Bool usbWcmInit(InputInfoPtr pInfo, char* id, size_t id_len, float *versi
 	usbdata = common->private;
 	*version = 0.0;
 
-	/* fetch vendor, product, and model name */
-	ioctl(pInfo->fd, EVIOCGID, &sID);
-	ioctl(pInfo->fd, EVIOCGNAME(id_len), id);
-
 	for (i = 0; i < ARRAY_SIZE(WacomModelDesc); i++)
 	{
 		if (sID.vendor == WacomModelDesc[i].vendor_id &&
@@ -546,6 +558,11 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 
 	if (!ISBITSET(ev,EV_ABS))
 	{
+		/* may be an expresskey only interface */
+		if (ISBITSET(common->wcmKeys, BTN_FORWARD) ||
+		    ISBITSET(common->wcmKeys, BTN_0))
+			goto pad_init;
+
 		xf86Msg(X_ERROR, "%s: no abs bits.\n", pInfo->name);
 		return !Success;
 	}
@@ -560,6 +577,11 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 	/* max x */
 	if (ioctl(pInfo->fd, EVIOCGABS(ABS_X), &absinfo) < 0)
 	{
+		/* may be a PAD only interface */
+		if (ISBITSET(common->wcmKeys, BTN_FORWARD) ||
+		    ISBITSET(common->wcmKeys, BTN_0))
+			goto pad_init;
+
 		xf86Msg(X_ERROR, "%s: unable to ioctl xmax value.\n", pInfo->name);
 		return !Success;
 	}
@@ -638,6 +660,16 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 			common->wcmMaxStripX = absinfo.maximum;
 	}
 
+	/* max touchring value for standalone pad tools */
+	common->wcmMinRing = 0;
+	common->wcmMaxRing = 71;
+	if (!ISBITSET(ev,EV_MSC) && ISBITSET(abs, ABS_WHEEL) &&
+			!ioctl(pInfo->fd, EVIOCGABS(ABS_WHEEL), &absinfo))
+	{
+		common->wcmMinRing = absinfo.minimum;
+		common->wcmMaxRing = absinfo.maximum;
+	}
+
 	/* X tilt range */
 	if (ISBITSET(abs, ABS_TILT_X) &&
 			!ioctl(pInfo->fd, EVIOCGABS(ABS_TILT_X), &absinfo))
@@ -755,8 +787,7 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 
 	if (ioctl(pInfo->fd, EVIOCGBIT(EV_SW, sizeof(sw)), sw) < 0)
 	{
-		xf86Msg(X_ERROR, "%s: usbProbeKeys unable to ioctl "
-			"sw bits.\n", pInfo->name);
+		xf86Msg(X_ERROR, "%s: unable to ioctl sw bits.\n", pInfo->name);
 		return 0;
 	}
 	else if (ISBITSET(sw, SW_MUTE_DEVICE))
@@ -765,7 +796,8 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 
 		memset(sw, 0, sizeof(sw));
 
-		ioctl(pInfo->fd, EVIOCGSW(sizeof(sw)), sw);
+		if (ioctl(pInfo->fd, EVIOCGSW(sizeof(sw)), sw) < 0)
+			xf86Msg(X_ERROR, "%s: unable to ioctl sw state.\n", pInfo->name);
 
 		if (ISBITSET(sw, SW_MUTE_DEVICE))
 			common->wcmHWTouchSwitchState = 0;
@@ -773,6 +805,7 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 			common->wcmHWTouchSwitchState = 1;
 	}
 
+pad_init:
 	usbWcmInitPadState(pInfo);
 
 	return Success;
@@ -800,11 +833,13 @@ static int usbParse(InputInfoPtr pInfo, const unsigned char* data, int len)
 {
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
+	struct input_event event;
 
 	if (len < sizeof(struct input_event))
 		return 0;
 
-	usbParseEvent(pInfo, (const struct input_event*)data);
+	memcpy(&event, data, sizeof(event));
+	usbParseEvent(pInfo, &event);
 	return common->wcmPktLength;
 }
 
@@ -1250,6 +1285,8 @@ static void usbParseAbsMTEvent(WacomCommonPtr common, struct input_event *event)
 			if (event->value >= 0) {
 				int serial = event->value + 1;
 				private->wcmMTChannel = usbChooseChannel(common, TOUCH_ID, serial);
+				if (private->wcmMTChannel < 0)
+					return;
 				ds = &common->wcmChannel[private->wcmMTChannel].work;
 				ds->serial_num = serial;
 			}
@@ -1782,10 +1819,10 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 	}
 
 	/* verify we have minimal data when entering prox */
-	if (ds->proximity && !dslast.proximity) {
+	if (ds->proximity && !dslast.proximity && ds->device_type != PAD_ID) {
 		struct input_absinfo absinfo;
 
-		if (!ds->x && !IsPad(priv)) {
+		if (!ds->x) {
 			if (ioctl(priv->pInfo->fd, EVIOCGABS(ABS_X), &absinfo) < 0)
 			{
 				DBG(-1, common, "unable to ioctl current x value.\n");
@@ -1793,7 +1830,7 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 			}
 			ds->x = absinfo.value;
 		}
-		if (!ds->y && !IsPad(priv)) {
+		if (!ds->y) {
 			if (ioctl(priv->pInfo->fd, EVIOCGABS(ABS_Y), &absinfo) < 0)
 			{
 				DBG(-1, common, "unable to ioctl current x value.\n");
