@@ -1,5 +1,5 @@
 /*
- * Copyright 2009 - 2010 by Ping Cheng, Wacom. <pingc@wacom.com>
+ * Copyright 2009 - 2013 by Ping Cheng, Wacom. <pingc@wacom.com>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -33,35 +33,37 @@
 static Bool wcmCheckSource(InputInfoPtr pInfo, dev_t min_maj)
 {
 	int match = 0;
-	char* device;
-	char* fsource = xf86CheckStrOption(pInfo->options, "_source", "");
 	InputInfoPtr pDevices = xf86FirstLocalDevice();
-	WacomCommonPtr pCommon = NULL;
-	char* psource;
 
-	for (; pDevices != NULL; pDevices = pDevices->next)
+	for (; !match && pDevices != NULL; pDevices = pDevices->next)
 	{
-		device = xf86CheckStrOption(pDevices->options, "Device", NULL);
+		char* device = xf86CheckStrOption(pDevices->options, "Device", NULL);
 
 		/* device can be NULL on some distros */
-		if (!device || !strstr(pDevices->drv->driverName, "wacom"))
+		if (!device)
+			continue;
+
+		free(device);
+
+		if (!strstr(pDevices->drv->driverName, "wacom"))
 			continue;
 
 		if (pInfo != pDevices)
 		{
-			psource = xf86CheckStrOption(pDevices->options, "_source", "");
-			pCommon = ((WacomDevicePtr)pDevices->private)->common;
+			WacomCommonPtr pCommon = ((WacomDevicePtr)pDevices->private)->common;
+			char* fsource = xf86CheckStrOption(pInfo->options, "_source", "");
+			char* psource = xf86CheckStrOption(pDevices->options, "_source", "");
+
 			if (pCommon->min_maj &&
 				pCommon->min_maj == min_maj)
 			{
 				/* only add the new tool if the matching major/minor
 				* was from the same source */
 				if (strcmp(fsource, psource))
-				{
 					match = 1;
-					break;
-				}
 			}
+			free(fsource);
+			free(psource);
 		}
 	}
 	if (match)
@@ -78,14 +80,14 @@ static Bool wcmCheckSource(InputInfoPtr pInfo, dev_t min_maj)
  * the xorg.conf and is then hotplugged through the server backend (HAL,
  * udev). In this case, the hotplugged one fails.
  */
-int wcmIsDuplicate(char* device, InputInfoPtr pInfo)
+int wcmIsDuplicate(const char* device, InputInfoPtr pInfo)
 {
 	struct stat st;
 	int isInUse = 0;
-	char* lsource = xf86CheckStrOption(pInfo->options, "_source", "");
+	char* lsource = xf86CheckStrOption(pInfo->options, "_source", NULL);
 
 	/* always allow xorg.conf defined tools to be added */
-	if (!strlen(lsource)) goto ret;
+	if (!lsource || !strlen(lsource)) goto ret;
 
 	if (stat(device, &st) == -1)
 	{
@@ -114,6 +116,7 @@ int wcmIsDuplicate(char* device, InputInfoPtr pInfo)
 		isInUse = 4;
 	}
 ret:
+	free(lsource);
 	return isInUse;
 }
 
@@ -136,7 +139,7 @@ Bool wcmIsAValidType(InputInfoPtr pInfo, const char* type)
 	int j, k, ret = FALSE;
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
-	char* dsource = xf86CheckStrOption(pInfo->options, "_source", "");
+	char* dsource;
 
 	if (!type)
 	{
@@ -144,6 +147,8 @@ Bool wcmIsAValidType(InputInfoPtr pInfo, const char* type)
 		return FALSE;
 	}
 
+	dsource = xf86CheckStrOption(pInfo->options, "_source", NULL);
+
 	/* walkthrough all types */
 	for (j = 0; j < ARRAY_SIZE(wcmType); j++)
 	{
@@ -163,7 +168,7 @@ Bool wcmIsAValidType(InputInfoPtr pInfo, const char* type)
 						    ret = FALSE;
 					}
 				}
-				else if (!strlen(dsource)) /* an user defined type */
+				else if (!dsource || !strlen(dsource)) /* an user defined type */
 				{
 					/* assume it is a valid type */
 					SETBIT(common->wcmKeys, wcmType[j].tool[k]);
@@ -177,6 +182,7 @@ Bool wcmIsAValidType(InputInfoPtr pInfo, const char* type)
 		xf86Msg(X_ERROR, "%s: Invalid type '%s' for this device.\n",
 			pInfo->name, type);
 
+	free(dsource);
 	return ret;
 }
 
@@ -191,6 +197,19 @@ int wcmDeviceTypeKeys(InputInfoPtr pInfo)
 
 	switch (priv->common->tablet_id)
 	{
+		case 0xF8:  /* Cintiq 24HDT */
+		case 0xF4:  /* Cintiq 24HD */
+			TabletSetFeature(priv->common, WCM_DUALRING | WCM_LCD);
+			/* fall through */
+
+		case 0x314: /* Intuos Pro S */
+		case 0x315: /* Intuos Pro M */
+		case 0x317: /* Intuos Pro L */
+		case 0x26:  /* I5 */
+		case 0x27:  /* I5 */
+		case 0x28:  /* I5 */
+		case 0x29:  /* I5 */
+		case 0x2A:  /* I5 */
 		case 0xB8:  /* I4 */
 		case 0xB9:  /* I4 */
 		case 0xBA:  /* I4 */
@@ -231,6 +250,7 @@ int wcmDeviceTypeKeys(InputInfoPtr pInfo)
 		case 0x37:  /* PL700 */
 		case 0x38:  /* PL510 */
 		case 0x39:  /* PL710 */
+		case 0x3A:  /* DTI520 */
 		case 0xC0:  /* DTF720 */
 		case 0xC2:  /* DTF720a */
 		case 0xC4:  /* DTF521 */
@@ -245,6 +265,8 @@ int wcmDeviceTypeKeys(InputInfoPtr pInfo)
 		case 0xC5:  /* CintiqV5 */
 		case 0xC6:  /* CintiqV5 */
 		case 0xCC:  /* CinitqV5 */
+		case 0xFA:  /* Cintiq 22HD */
+		case 0x5B:  /* Cintiq 22HDT Pen */
 			TabletSetFeature(priv->common, WCM_LCD);
 			/* fall through */
 		case 0xB0:  /* I3 */
@@ -257,24 +279,58 @@ int wcmDeviceTypeKeys(InputInfoPtr pInfo)
 			TabletSetFeature(priv->common, WCM_STRIP | WCM_ROTATION);
 			break;
 
+		case 0x100: /* TPC with MT */
+		case 0x101: /* TPC with MT */
+		case 0x10D: /* TPC with MT */
+		case 0x116: /* TPC with 1FGT */
+		case 0x12C: /* TPC */
+		case 0x4001: /* TPC with MT */
+		case 0x4004: /* TPC with MT (no pen on Motion) */
+		case 0x5000: /* TPC with MT */
+		case 0x5002: /* TPC with MT */
 		case 0xE2: /* TPC with 2FGT */
 		case 0xE3: /* TPC with 2FGT */
+		case 0xE5: /* TPC with MT */
 		case 0xE6: /* TPC with 2FGT */
-			TabletSetFeature(priv->common, WCM_TPC);
-			break;
-
 		case 0x93: /* TPC with 1FGT */
 		case 0x9A: /* TPC with 1FGT */
+		case 0xEC: /* TPC with 1FGT */
+		case 0xED: /* TPC with 1FGT */
 		case 0x90: /* TPC */
 		case 0x97: /* TPC */
+		case 0xEF: /* TPC */
 			TabletSetFeature(priv->common, WCM_TPC);
 			break;
 
 		case 0x9F:
+		case 0xF6: /* Cintiq 24HDT Touch */
+		case 0x57: /* DTK2241 */
+		case 0x59: /* DTH2242 Pen */
+		case 0x5D: /* DTH2242 Touch */
+		case 0x5E: /* Cintiq 22HDT Touch */
+		case 0x304:/* Cintiq 13HD */
+		case 0x34A:/* MobileStudio Pro 13 Touch */
+		case 0x34B:/* MobileStudio Pro 16 Touch */
+		case 0x34D:/* MobileStudio Pro 13 */
+		case 0x34E:/* MobileStudio Pro 13 */
+		case 0x34F:/* Cintiq Pro 13 FHD */
+		case 0x350:/* Cintiq Pro 16 UHD */
+		case 0x353:/* Cintiq Pro 13 FHD Touch */
+		case 0x354:/* Cintiq Pro 13 UHD Touch */
 			TabletSetFeature(priv->common, WCM_LCD);
 			break;
 	}
 
+#ifdef INPUT_PROP_DIRECT
+	{
+		int rc;
+		unsigned long prop[NBITS(INPUT_PROP_MAX)] = {0};
+
+		rc = ioctl(pInfo->fd, EVIOCGPROP(sizeof(prop)), prop);
+		if (rc >= 0 && ISBITSET(prop, INPUT_PROP_DIRECT))
+			TabletSetFeature(priv->common, WCM_LCD);
+	}
+#endif
 	if (ISBITSET(common->wcmKeys, BTN_TOOL_PEN))
 		TabletSetFeature(priv->common, WCM_PEN);
 
@@ -315,8 +371,8 @@ input_option_new(InputOption *list, char *key, char *value)
 	InputOption *new;
 
 	new = calloc(1, sizeof(InputOption));
-	new->key = key;
-	new->value = value;
+	new->key = strdup(key);
+	new->value = strdup(value);
 	new->next = list;
 	return new;
 }
@@ -352,7 +408,7 @@ static InputOption *wcmOptionDupConvert(InputInfoPtr pInfo, const char* basename
 	WacomToolPtr ser = common->serials;
 	InputOption *iopts = NULL;
 	char *name;
-	pointer options;
+	pointer options, o;
 	int rc;
 
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
@@ -390,13 +446,15 @@ static InputOption *wcmOptionDupConvert(InputInfoPtr pInfo, const char* basename
 
 	free(name);
 
-	while(options)
+	o = options;
+	while(o)
 	{
 		iopts = input_option_new(iopts,
-					 xf86OptionName(options),
-					 xf86OptionValue(options));
-		options = xf86NextOption(options);
+					 xf86OptionName(o),
+					 xf86OptionValue(o));
+		o = xf86NextOption(o);
 	}
+	xf86OptionListFree(options);
 	return iopts;
 }
 
@@ -412,10 +470,12 @@ static InputAttributes* wcmDuplicateAttributes(InputInfoPtr pInfo,
 {
 	int rc;
 	InputAttributes *attr;
+	char *product;
+
 	attr = DuplicateInputAttributes(pInfo->attrs);
-	rc = asprintf(&attr->product, "%s %s", attr->product, type);
-	if (rc == -1)
-		attr->product = NULL;
+	rc = asprintf(&product, "%s %s", attr->product, type);
+	free(attr->product);
+	attr->product = (rc != -1) ? product : NULL;
 	return attr;
 }
 #endif
@@ -446,11 +506,19 @@ wcmHotplugDevice(ClientPtr client, pointer closure )
 	WacomHotplugInfo *hotplug_info = closure;
 	DeviceIntPtr dev; /* dummy */
 
+#if HAVE_THREADED_INPUT
+	input_lock();
+#endif
+
 	NewInputDeviceRequest(hotplug_info->input_options,
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 9
 			      hotplug_info->attrs,
 #endif
 			      &dev);
+#if HAVE_THREADED_INPUT
+	input_unlock();
+#endif
+
 	input_option_free_list(&hotplug_info->input_options);
 
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 11
@@ -469,7 +537,7 @@ wcmHotplugDevice(ClientPtr client, pointer closure )
  *
  * Note that we don't actually hotplug the device here. We store the
  * information needed to hotplug the device later and then queue the
- * hotplug. The server will come back and call the @wcmHotplugDevice
+ * hotplug. The server will come back and call the @ref wcmHotplugDevice
  * later.
  *
  * @param pInfo The parent device
@@ -560,16 +628,20 @@ void wcmHotplugOthers(InputInfoPtr pInfo, const char *basename)
  * This changes the source to _driver/wacom, all auto-hotplugged devices
  * will have the same source.
  */
-int wcmNeedAutoHotplug(InputInfoPtr pInfo, const char **type)
+int wcmNeedAutoHotplug(InputInfoPtr pInfo, char **type)
 {
-	char *source = xf86CheckStrOption(pInfo->options, "_source", "");
+	char *source = xf86CheckStrOption(pInfo->options, "_source", NULL);
 	int i;
+	int rc = 0;
 
 	if (*type) /* type specified, don't hotplug */
-		return 0;
+		goto out;
 
-	if (strcmp(source, "server/hal") && strcmp(source, "server/udev"))
-		return 0;
+	if (!source) /* xorg.conf device, don't auto-pick type */
+		goto out;
+
+	if (source && strcmp(source, "server/hal") && strcmp(source, "server/udev"))
+		goto out;
 
 	/* no type specified, so we need to pick the first one applicable
 	 * for our device */
@@ -577,13 +649,14 @@ int wcmNeedAutoHotplug(InputInfoPtr pInfo, const char **type)
 	{
 		if (wcmIsAValidType(pInfo, wcmType[i].type))
 		{
+			free(*type);
 			*type = strdup(wcmType[i].type);
 			break;
 		}
 	}
 
 	if (!*type)
-		return 0;
+		goto out;
 
 	xf86Msg(X_INFO, "%s: type not specified, assuming '%s'.\n", pInfo->name, *type);
 	xf86Msg(X_INFO, "%s: other types will be automatically added.\n", pInfo->name);
@@ -592,7 +665,11 @@ int wcmNeedAutoHotplug(InputInfoPtr pInfo, const char **type)
 	pInfo->options = xf86AddNewOption(pInfo->options, "Type", *type);
 	pInfo->options = xf86ReplaceStrOption(pInfo->options, "_source", "_driver/wacom");
 
-	return 1;
+	rc = 1;
+
+	free(source);
+out:
+	return rc;
 }
 
 int wcmParseSerials (InputInfoPtr pInfo)
@@ -626,6 +703,7 @@ int wcmParseSerials (InputInfoPtr pInfo)
 			{
 				xf86Msg(X_ERROR, "%s: %s is invalid serial string.\n",
 					pInfo->name, tok);
+				free(ser);
 				return 1;
 			}
 
@@ -659,7 +737,7 @@ int wcmParseSerials (InputInfoPtr pInfo)
 					pInfo->name, serial, name);
 				ser->name = strdup(name);
 			}
-			else ser->name = ""; /*no name yet*/
+			else ser->name = strdup(""); /*no name yet*/
 
 			if (common->serials == NULL)
 				common->serials = ser;
@@ -678,7 +756,9 @@ int wcmParseSerials (InputInfoPtr pInfo)
 }
 
 /**
- * Parse the options for this device.
+ * Parse the pre-init options for this device. Most useful for options
+ * needed to properly init a device (baud rate for example).
+ *
  * Note that parameters is_primary and is_dependent are mutually exclusive,
  * though both may be false in the case of an xorg.conf device.
  *
@@ -686,13 +766,14 @@ int wcmParseSerials (InputInfoPtr pInfo)
  * hotplugging, False if the device is a depent or xorg.conf device.
  * @param is_hotplugged True if the device is a dependent device, FALSE
  * otherwise.
- * @retvalue True on success or False otherwise.
+ * @retval True on success or False otherwise.
  */
-Bool wcmParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent)
+Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
+			    Bool is_dependent)
 {
 	WacomDevicePtr  priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr  common = priv->common;
-	char            *s, b[12];
+	char            *s;
 	int		i;
 	WacomToolPtr    tool = NULL;
 	int		tpc_button_is_on;
@@ -715,6 +796,8 @@ Bool wcmParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent)
 		 */
 	}
 
+	free(s);
+
 	/* Pad is always in absolute mode.
 	 * The pad also defaults to wheel scrolling, unlike the pens
 	 * (interesting effects happen on ArtPen and others with build-in
@@ -722,8 +805,8 @@ Bool wcmParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent)
 	 */
 	if (IsPad(priv))
 	{
-		priv->wheelup = 4;
-		priv->wheeldn = 5;
+		priv->wheel_default[WHEEL_ABS_UP] = priv->wheel_default[WHEEL2_ABS_UP] = 4;
+		priv->wheel_default[WHEEL_ABS_DN] = priv->wheel_default[WHEEL2_ABS_DN] = 5;
 		set_absolute(pInfo, TRUE);
 	}
 
@@ -751,6 +834,7 @@ Bool wcmParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent)
 					" device\n", pInfo->name);
 		else
 			wcmRotateTablet(pInfo, rotation);
+		free(s);
 	}
 
 	common->wcmRawSample = xf86SetIntOption(pInfo->options, "RawSample",
@@ -772,10 +856,6 @@ Bool wcmParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent)
 			common->wcmSuppress = DEFAULT_SUPPRESS;
 	}
 
-	if (xf86SetBoolOption(pInfo->options, "Tilt",
-			(common->wcmFlags & TILT_REQUEST_FLAG)))
-		common->wcmFlags |= TILT_REQUEST_FLAG;
-
 	/* pressure curve takes control points x1,y1,x2,y2
 	 * values in range from 0..100.
 	 * Linear curve is 0,0,100,100
@@ -793,6 +873,12 @@ Bool wcmParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent)
 		else
 			wcmSetPressureCurve(priv,a,b,c,d);
 	}
+	free(s);
+
+	if (xf86SetBoolOption(pInfo->options, "Pressure2K", 0)) {
+		xf86Msg(X_CONFIG, "%s: Using 2K pressure levels\n", pInfo->name);
+		priv->maxCurve = 2048;
+	}
 
 	/*Serials of tools we want hotpluged*/
 	if (wcmParseSerials (pInfo) != 0)
@@ -829,7 +915,7 @@ Bool wcmParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent)
 
 		if(toollist) /* Already have a tool with the same type/serial */
 		{
-			xf86Msg(X_ERROR, "%s: already have a tool with type/serial %d/%d.",
+			xf86Msg(X_ERROR, "%s: already have a tool with type/serial %d/%d.\n",
 					pInfo->name, tool->typeid, tool->serial);
 			goto error;
 		} else /* No match on existing tool/serial, add tool to the end of the list */
@@ -844,8 +930,6 @@ Bool wcmParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent)
 	common->wcmThreshold = xf86SetIntOption(pInfo->options, "Threshold",
 			common->wcmThreshold);
 
-	common->wcmMaxZ = xf86SetIntOption(pInfo->options, "MaxZ",
-					   common->wcmMaxZ);
 	if (xf86SetBoolOption(pInfo->options, "ButtonsOnly", 0))
 		priv->flags |= BUTTONS_ONLY_FLAG;
 
@@ -877,6 +961,11 @@ Bool wcmParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent)
 		else if (touch_is_on != common->wcmTouch)
 			xf86Msg(X_WARNING, "%s: Touch option can only be set "
 				"by a touch tool.\n", pInfo->name);
+
+		if (TabletHasFeature(common, WCM_1FGT))
+			common->wcmMaxContacts = 1;
+		else
+			common->wcmMaxContacts = 2;
 	}
 
 	/* 2FG touch device */
@@ -897,30 +986,29 @@ Bool wcmParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent)
 			xf86Msg(X_WARNING, "%s: Touch gesture option can only "
 				"be set by a touch tool.\n", pInfo->name);
 
-		common->wcmGestureParameters.wcmZoomDistance =
-			xf86SetIntOption(pInfo->options, "ZoomDistance",
-			common->wcmGestureParameters.wcmZoomDistanceDefault);
-
-		common->wcmGestureParameters.wcmScrollDistance =
-			xf86SetIntOption(pInfo->options, "ScrollDistance",
-			common->wcmGestureParameters.wcmScrollDistanceDefault);
-
 		common->wcmGestureParameters.wcmTapTime =
 			xf86SetIntOption(pInfo->options, "TapTime",
-			common->wcmGestureParameters.wcmTapTimeDefault);
+			common->wcmGestureParameters.wcmTapTime);
+	}
+
+	if (IsStylus(priv) || IsEraser(priv)) {
+		common->wcmPressureRecalibration
+			= xf86SetBoolOption(pInfo->options,
+					    "PressureRecalibration", 1);
 	}
 
 	/* Swap stylus buttons 2 and 3 for Tablet PCs */
 	if (TabletHasFeature(common, WCM_TPC) && IsStylus(priv))
 	{
-		priv->button[1] = 3;
-		priv->button[2] = 2;
+		priv->button_default[1] = 3;
+		priv->button_default[2] = 2;
 	}
 
 	for (i=0; i<WCM_MAX_BUTTONS; i++)
 	{
+		char b[12];
 		sprintf(b, "Button%d", i+1);
-		priv->button[i] = xf86SetIntOption(pInfo->options, b, priv->button[i]);
+		priv->button_default[i] = xf86SetIntOption(pInfo->options, b, priv->button_default[i]);
 	}
 
 	/* Now parse class-specific options */
@@ -930,8 +1018,60 @@ Bool wcmParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent)
 
 	return TRUE;
 error:
-	free(tool);
 	return FALSE;
 }
 
+/* The values were based on trail and error. */
+#define WCM_BAMBOO3_MAXX 4096.0
+#define WCM_BAMBOO3_ZOOM_DISTANCE 180.0
+#define WCM_BAMBOO3_SCROLL_DISTANCE 80.0
+#define WCM_BAMBOO3_SCROLL_SPREAD_DISTANCE 350.0
+
+/**
+ * Parse post-init options for this device. Useful for overriding HW
+ * specific options computed during init phase (HW distances for example).
+ *
+ * Note that parameters is_primary and is_dependent are mutually exclusive,
+ * though both may be false in the case of an xorg.conf device.
+ *
+ * @param is_primary True if the device is the parent device for
+ * hotplugging, False if the device is a depent or xorg.conf device.
+ * @param is_hotplugged True if the device is a dependent device, FALSE
+ * otherwise.
+ * @retval True on success or False otherwise.
+ */
+Bool wcmPostInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
+			     Bool is_dependent)
+{
+	WacomDevicePtr  priv = (WacomDevicePtr)pInfo->private;
+	WacomCommonPtr  common = priv->common;
+
+	common->wcmMaxZ = xf86SetIntOption(pInfo->options, "MaxZ",
+					   common->wcmMaxZ);
+
+	/* 2FG touch device */
+	if (TabletHasFeature(common, WCM_2FGT) && IsTouch(priv))
+	{
+		int zoom_distance = common->wcmMaxTouchX *
+			(WCM_BAMBOO3_ZOOM_DISTANCE / WCM_BAMBOO3_MAXX);
+		int scroll_distance = common->wcmMaxTouchX *
+			(WCM_BAMBOO3_SCROLL_DISTANCE / WCM_BAMBOO3_MAXX);
+
+		common->wcmGestureParameters.wcmZoomDistance =
+			xf86SetIntOption(pInfo->options, "ZoomDistance",
+					 zoom_distance);
+
+		common->wcmGestureParameters.wcmScrollDistance =
+			xf86SetIntOption(pInfo->options, "ScrollDistance",
+					 scroll_distance);
+
+		common->wcmGestureParameters.wcmMaxScrollFingerSpread =
+			common->wcmMaxTouchX *
+			(WCM_BAMBOO3_SCROLL_SPREAD_DISTANCE / WCM_BAMBOO3_MAXX);
+	}
+
+
+	return TRUE;
+}
+
 /* vim: set noexpandtab tabstop=8 shiftwidth=8: */
