@@ -79,29 +79,103 @@ int wcmDevSwitchMode(ClientPtr client, DeviceIntPtr dev, int mode)
 	return wcmDevSwitchModeCall(pInfo, mode);
 }
 
-Atom prop_devnode;
-Atom prop_rotation;
-Atom prop_tablet_area;
-Atom prop_pressurecurve;
-Atom prop_serials;
-Atom prop_serial_binding;
-Atom prop_strip_buttons;
-Atom prop_wheel_buttons;
-Atom prop_tv_resolutions;
-Atom prop_cursorprox;
-Atom prop_threshold;
-Atom prop_suppress;
-Atom prop_touch;
-Atom prop_gesture;
-Atom prop_gesture_param;
-Atom prop_hover;
-Atom prop_tooltype;
-Atom prop_btnactions;
-Atom prop_product_id;
+static Atom prop_devnode;
+static Atom prop_rotation;
+static Atom prop_tablet_area;
+static Atom prop_pressurecurve;
+static Atom prop_serials;
+static Atom prop_serial_binding;
+static Atom prop_strip_buttons;
+static Atom prop_wheel_buttons;
+static Atom prop_cursorprox;
+static Atom prop_threshold;
+static Atom prop_suppress;
+static Atom prop_touch;
+static Atom prop_hardware_touch;
+static Atom prop_gesture;
+static Atom prop_gesture_param;
+static Atom prop_hover;
+static Atom prop_tooltype;
+static Atom prop_btnactions;
+static Atom prop_product_id;
+static Atom prop_pressure_recal;
 #ifdef DEBUG
-Atom prop_debuglevels;
+static Atom prop_debuglevels;
 #endif
 
+/**
+ * Calculate a user-visible pressure level from a driver-internal pressure
+ * level. Pressure settings exposed to the user assume a range of 0-2047
+ * while the driver scales everything to a range of 0-maxCurve.
+ */
+static inline int wcmInternalToUserPressure(InputInfoPtr pInfo, int pressure)
+{
+	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
+	return pressure / (priv->maxCurve / 2048);
+}
+
+/**
+ * Calculate a driver-internal pressure level from a user-visible pressure
+ * level. Pressure settings exposed to the user assume a range of 0-2047
+ * while the driver scales everything to a range of 0-maxCurve.
+ */
+static inline int wcmUserToInternalPressure(InputInfoPtr pInfo, int pressure)
+{
+	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
+	return pressure * (priv->maxCurve / 2048);
+}
+
+/**
+ * Resets an arbitrary Action property, given a pointer to the old
+ * handler and information about the new Action.
+ */
+static void wcmResetAction(InputInfoPtr pInfo, const char *name, int index,
+                           Atom *handler, unsigned int (*action)[256],
+                           unsigned int (*new_action)[256], Atom prop, int nprop)
+{
+	handler[index] = MakeAtom(name, strlen(name), TRUE);
+	memset(action[index], 0, sizeof(action[index]));
+	memcpy(action[index], *new_action, sizeof(*new_action));
+	XIChangeDeviceProperty(pInfo->dev, handler[index], XA_INTEGER, 32,
+			       PropModeReplace, 1, (char*)new_action, FALSE);
+}
+
+static void wcmResetButtonAction(InputInfoPtr pInfo, int button, int nbuttons)
+{
+	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
+	unsigned int new_action[256] = {};
+	int x11_button = priv->button_default[button];
+	char name[64];
+
+	sprintf(name, "Wacom button action %d", button);
+	new_action[0] = AC_BUTTON | AC_KEYBTNPRESS | x11_button;
+	wcmResetAction(pInfo, name, button, priv->btn_actions, priv->keys, &new_action, prop_btnactions, nbuttons);
+}
+
+static void wcmResetStripAction(InputInfoPtr pInfo, int index)
+{
+	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
+	unsigned int new_action[256] = {};
+	char name[64];
+
+	sprintf(name, "Wacom strip action %d", index);
+	new_action[0] =	AC_BUTTON | AC_KEYBTNPRESS | (priv->strip_default[index]);
+	new_action[1] = AC_BUTTON | (priv->strip_default[index]);
+	wcmResetAction(pInfo, name, index, priv->strip_actions, priv->strip_keys, &new_action, prop_strip_buttons, 4);
+}
+
+static void wcmResetWheelAction(InputInfoPtr pInfo, int index)
+{
+	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
+	unsigned int new_action[256] = {};
+	char name[64];
+
+	sprintf(name, "Wacom wheel action %d", index);
+	new_action[0] = AC_BUTTON | AC_KEYBTNPRESS | (priv->wheel_default[index]);
+	new_action[1] = AC_BUTTON | (priv->wheel_default[index]);
+	wcmResetAction(pInfo, name, index, priv->wheel_actions, priv->wheel_keys, &new_action, prop_wheel_buttons, 6);
+}
+
 /**
  * Registers a property for the input device. This function registers
  * the property name atom, as well as creates the property itself.
@@ -118,7 +192,7 @@ Atom prop_debuglevels;
  * @param values Pointer to 32 bit integer array of initial property values
  * @return Atom handle of property name
  */
-static Atom InitWcmAtom(DeviceIntPtr dev, char *name, Atom type, int format, int nvalues, int *values)
+static Atom InitWcmAtom(DeviceIntPtr dev, const char *name, Atom type, int format, int nvalues, int *values)
 {
 	int i;
 	Atom atom;
@@ -157,7 +231,7 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
 	WacomCommonPtr common = priv->common;
 	int values[WCM_MAX_BUTTONS];
-	int nbuttons;
+	int i;
 
 	DBG(10, priv, "\n");
 
@@ -176,7 +250,9 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 	}
 
 	values[0] = common->wcmRotate;
-	prop_rotation = InitWcmAtom(pInfo->dev, WACOM_PROP_ROTATION, XA_INTEGER, 8, 1, values);
+	if (!IsPad(priv)) {
+		prop_rotation = InitWcmAtom(pInfo->dev, WACOM_PROP_ROTATION, XA_INTEGER, 8, 1, values);
+	}
 
 	if (IsPen(priv) || IsTouch(priv)) {
 		values[0] = priv->nPressCtrl[0];
@@ -187,8 +263,8 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 	}
 
 	values[0] = common->tablet_id;
-	values[1] = priv->old_serial;
-	values[2] = priv->old_device_id;
+	values[1] = priv->oldState.serial_num;
+	values[2] = priv->oldState.device_id;
 	values[3] = priv->cur_serial;
 	values[4] = priv->cur_device_id;
 	prop_serials = InitWcmAtom(pInfo->dev, WACOM_PROP_SERIALIDS, XA_INTEGER, 32, 5, values);
@@ -202,6 +278,7 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 	}
 
 	values[0] = (!common->wcmMaxZ) ? 0 : common->wcmThreshold;
+	values[0] = wcmInternalToUserPressure(pInfo, values[0]);
 	prop_threshold = InitWcmAtom(pInfo->dev, WACOM_PROP_PRESSURE_THRESHOLD, XA_INTEGER, 32, 1, values);
 
 	values[0] = common->wcmSuppress;
@@ -211,6 +288,11 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 	values[0] = common->wcmTouch;
 	prop_touch = InitWcmAtom(pInfo->dev, WACOM_PROP_TOUCH, XA_INTEGER, 8, 1, values);
 
+	if (common->wcmHasHWTouchSwitch && IsTouch(priv)) {
+		values[0] = common->wcmHWTouchSwitchState;
+		prop_hardware_touch = InitWcmAtom(pInfo->dev, WACOM_PROP_HARDWARE_TOUCH, XA_INTEGER, 8, 1, values);
+	}
+
 	if (IsStylus(priv)) {
 		values[0] = !common->wcmTPCButton;
 		prop_hover = InitWcmAtom(pInfo->dev, WACOM_PROP_HOVER, XA_INTEGER, 8, 1, values);
@@ -227,21 +309,31 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 	values[0] = MakeAtom(pInfo->type_name, strlen(pInfo->type_name), TRUE);
 	prop_tooltype = InitWcmAtom(pInfo->dev, WACOM_PROP_TOOL_TYPE, XA_ATOM, 32, 1, values);
 
-
-	/* default to no actions */
-	nbuttons = min(max(priv->nbuttons + 4, 7), WCM_MAX_BUTTONS);
 	memset(values, 0, sizeof(values));
-	prop_btnactions = InitWcmAtom(pInfo->dev, WACOM_PROP_BUTTON_ACTIONS, XA_ATOM, 32, nbuttons, values);
+	prop_btnactions = InitWcmAtom(pInfo->dev, WACOM_PROP_BUTTON_ACTIONS, XA_ATOM, 32, priv->nbuttons, values);
+	for (i = 0; i < priv->nbuttons; i++)
+		wcmResetButtonAction(pInfo, i, priv->nbuttons);
 
 	if (IsPad(priv)) {
 		memset(values, 0, sizeof(values));
 		prop_strip_buttons = InitWcmAtom(pInfo->dev, WACOM_PROP_STRIPBUTTONS, XA_ATOM, 32, 4, values);
+		for (i = 0; i < 4; i++)
+			wcmResetStripAction(pInfo, i);
 	}
 
 	if (IsPad(priv) || IsCursor(priv))
 	{
 		memset(values, 0, sizeof(values));
 		prop_wheel_buttons = InitWcmAtom(pInfo->dev, WACOM_PROP_WHEELBUTTONS, XA_ATOM, 32, 6, values);
+		for (i = 0; i < 6; i++)
+			wcmResetWheelAction(pInfo, i);
+	}
+
+	if (IsStylus(priv) || IsEraser(priv)) {
+		values[0] = common->wcmPressureRecalibration;
+		prop_pressure_recal = InitWcmAtom(pInfo->dev,
+						  WACOM_PROP_PRESSURE_RECAL,
+						  XA_INTEGER, 8, 1, values);
 	}
 
 	values[0] = common->vendor_id;
@@ -264,21 +356,83 @@ static int wcmFindProp(Atom property, Atom *prop_list, int nprops)
 	/* check all properties used for button actions */
 	for (i = 0; i < nprops; i++)
 		if (prop_list[i] == property)
-			break;
+			return i;
 
-	if (i >= nprops)
-		return -BadAtom;
-
-	return i;
+	return -BadAtom;
 }
 
-static int wcmSanityCheckProperty(XIPropertyValuePtr prop)
+/**
+ * Obtain a pointer to the the handler and action list for a given Action
+ * property. This function searches the button, wheel, and strip property
+ * handler lists.
+ *
+ * @param priv          The device whose handler lists should be searched
+ * @param property      The Action property that should be searched for
+ * @param[out] handler  Returns a pointer to the property's handler
+ * @param[out] action   Returns a pointer to the property's action list
+ * @return              'true' if the property was found. Neither out parameter
+ *                      will be null if this is the case.
+ */
+static BOOL wcmFindActionHandler(WacomDevicePtr priv, Atom property, Atom **handler, unsigned int (**action)[256])
+{
+	int offset;
+
+	offset = wcmFindProp(property, priv->btn_actions, ARRAY_SIZE(priv->btn_actions));
+	if (offset >=0)
+	{
+		*handler = &priv->btn_actions[offset];
+		*action  = &priv->keys[offset];
+		return TRUE;
+	}
+
+	offset = wcmFindProp(property, priv->wheel_actions, ARRAY_SIZE(priv->wheel_actions));
+	if (offset >= 0)
+	{
+		*handler = &priv->wheel_actions[offset];
+		*action  = &priv->wheel_keys[offset];
+		return TRUE;
+	}
+
+	offset = wcmFindProp(property, priv->strip_actions, ARRAY_SIZE(priv->strip_actions));
+	if (offset >= 0)
+	{
+		*handler = &priv->strip_actions[offset];
+		*action  = &priv->strip_keys[offset];
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static int wcmCheckActionProperty(WacomDevicePtr priv, Atom property, XIPropertyValuePtr prop)
 {
 	CARD32 *data;
 	int j;
 
-	if (prop->size >= 255 || prop->format != 32 || prop->type != XA_INTEGER)
+	if (!property) {
+		DBG(3, priv, "ERROR: Atom is NONE\n");
 		return BadMatch;
+	}
+
+	if (prop == NULL) {
+		DBG(3, priv, "ERROR: Value is NULL\n");
+		return BadMatch;
+	}
+
+	if (prop->size >= 255) {
+		DBG(3, priv, "ERROR: Too many values (%ld > 255)\n", prop->size);
+		return BadMatch;
+	}
+
+	if (prop->format != 32) {
+		DBG(3, priv, "ERROR: Incorrect value format (%d != 32)\n", prop->format);
+		return BadMatch;
+	}
+
+	if (prop->type != XA_INTEGER) {
+		DBG(3, priv, "ERROR: Incorrect value type (%d != XA_INTEGER)\n", prop->type);
+		return BadMatch;
+	}
 
 	data = (CARD32*)prop->data;
 
@@ -287,18 +441,22 @@ static int wcmSanityCheckProperty(XIPropertyValuePtr prop)
 		int code = data[j] & AC_CODE;
 		int type = data[j] & AC_TYPE;
 
+		DBG(10, priv, "Index %d == %d (type: %d, code: %d)\n", j, data[j], type, code);
+
 		switch(type)
 		{
 			case AC_KEY:
 				break;
 			case AC_BUTTON:
-				if (code > WCM_MAX_BUTTONS)
+				if (code > WCM_MAX_X11BUTTON) {
+					DBG(3, priv, "ERROR: AC_BUTTON code too high (%d > %d)\n", code, WCM_MAX_X11BUTTON);
 					return BadValue;
+				}
 				break;
-			case AC_DISPLAYTOGGLE:
 			case AC_MODETOGGLE:
 				break;
 			default:
+				DBG(3, priv, "ERROR: Unknown command\n");
 				return BadValue;
 		}
 	}
@@ -307,99 +465,61 @@ static int wcmSanityCheckProperty(XIPropertyValuePtr prop)
 }
 
 /**
- * Store the new value of the property in one of the driver's internal
- * property handler lists. Properties stored there will be checked for value
- * changes whenever updated.
+ * An 'Action' property (such as an element of "Wacom Button Actions")
+ * defines an action to be performed for some event. The property is
+ * validated, and then saved for later use. Both the property itself
+ * (as 'handler') and the data it references (as 'action') are saved.
+ *
+ * @param dev        The device being modified
+ * @param property   The Action property being set
+ * @param prop       The data contained in 'property'
+ * @param checkonly  'true' if the property should only be checked for validity
+ * @param handler    Pointer to the handler that must be updated
+ * @param action     Pointer to the action list that must be updated
  */
-static void wcmUpdateActionPropHandlers(XIPropertyValuePtr prop, Atom *handlers)
-{
-	int i;
-	CARD32 *values = (CARD32*)prop->data;
-
-	/* any action property needs to be registered for this handler. */
-	for (i = 0; i < prop->size; i++)
-		handlers[i] = values[i];
-}
-
-static void wcmUpdateButtonKeyActions(DeviceIntPtr dev, XIPropertyValuePtr prop,
-					unsigned int (*keys)[256], int skeys)
-{
-	Atom *values = (Atom*)prop->data;
-	XIPropertyValuePtr val;
-	int i, j;
-
-	for (i = 0; i < prop->size; i++)
-	{
-		/* keys is one based array to align with X buttons */
-		memset(keys[i+1], 0, sizeof(keys[i+1]));
-
-		if (!values[i])
-			continue;
-
-		XIGetDeviceProperty(dev, values[i], &val);
-
-		for (j = 0; j < val->size; j++)
-			keys[i+1][j] = ((unsigned int*)val->data)[j];
-	}
-}
-
-/* Change the properties that hold the actual button actions */
-static int wcmSetActionProperties(DeviceIntPtr dev, Atom property,
-				  XIPropertyValuePtr prop, BOOL checkonly)
+static int wcmSetActionProperty(DeviceIntPtr dev, Atom property,
+				XIPropertyValuePtr prop, BOOL checkonly,
+				Atom *handler, unsigned int (*action)[256])
 {
 	InputInfoPtr pInfo = (InputInfoPtr) dev->public.devicePrivate;
 	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
-	int i;
-	int rc;
+	int rc, i;
 
+	DBG(5, priv, "%s new actions for Atom %d\n", checkonly ? "Checking" : "Setting", property);
 
-	DBG(10, priv, "\n");
-
-	rc = wcmSanityCheckProperty(prop);
-	if (rc != Success)
+	rc = wcmCheckActionProperty(priv, property, prop);
+	if (rc != Success) {
+		const char *msg = NULL;
+		switch (rc) {
+			case BadMatch: msg = "BadMatch"; break;
+			case BadValue: msg = "BadValue"; break;
+			default: msg = "UNKNOWN"; break;
+		}
+		DBG(3, priv, "Action validation failed with code %d (%s)\n", rc, msg);
 		return rc;
-
-	i = wcmFindProp(property, priv->btn_actions, ARRAY_SIZE(priv->btn_actions));
-	if (i >= 0)
-	{
-		if (!checkonly)
-		{
-			XIGetDeviceProperty(dev, prop_btnactions, &prop);
-			wcmUpdateButtonKeyActions(dev, prop, priv->keys, ARRAY_SIZE(priv->keys));
-		}
-	} else
-	{
-		i = wcmFindProp(property, priv->wheel_actions,
-					ARRAY_SIZE(priv->wheel_actions));
-		if (i >= 0) {
-			if (!checkonly)
-			{
-				XIGetDeviceProperty(dev, prop_wheel_buttons, &prop);
-				wcmUpdateButtonKeyActions(dev, prop,
-						priv->wheel_keys,
-						ARRAY_SIZE(priv->wheel_keys));
-			}
-		} else
-		{
-			i = wcmFindProp(property, priv->strip_actions, ARRAY_SIZE(priv->strip_actions));
-			if (i >= 0 && !checkonly)
-			{
-				XIGetDeviceProperty(dev, prop_strip_buttons, &prop);
-				wcmUpdateButtonKeyActions(dev, prop, priv->strip_keys, ARRAY_SIZE(priv->strip_keys));
-			}
-		}
 	}
 
-	return abs(i);
+	if (!checkonly)
+	{
+		memset(action, 0, sizeof(*action));
+		for (i = 0; i < prop->size; i++)
+			(*action)[i] = ((unsigned int*)prop->data)[i];
+		*handler = property;
+	}
+
+	return Success;
 }
 
-static int wcmCheckActionProp(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
+static int wcmCheckActionsProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
 {
 	InputInfoPtr pInfo = (InputInfoPtr) dev->public.devicePrivate;
 	XIPropertyValuePtr val;
 	Atom *values = (Atom*)prop->data;
 	int i;
 
+	if (prop->format != 32 || prop->type != XA_ATOM)
+		return BadMatch;
+
 	for (i = 0; i < prop->size; i++)
 	{
 		if (!values[i])
@@ -415,169 +535,79 @@ static int wcmCheckActionProp(DeviceIntPtr dev, Atom property, XIPropertyValuePt
 	return Success;
 }
 
-/* Change the property that refers to which properties the actual button
- * actions are stored in */
-static int wcmSetPropertyButtonActions(DeviceIntPtr dev, Atom property,
-				       XIPropertyValuePtr prop, BOOL checkonly)
+/**
+ * An 'Actions' property (such as "Wacom Button Actions") stores a list of
+ * 'Action' properties that define an action to be performed. This function
+ * goes through the list of actions and saves each one.
+ *
+ * @param dev        The device being modified
+ * @param property   The Actions property being set
+ * @param prop       The data contained in 'property'
+ * @param checkonly  'true' if the property should only be checked for validity
+ * @param size       Expected number of elements in 'prop'
+ * @param handlers   List of handlers that must be updated
+ * @param actions    List of actions that must be updated
+ */
+static int wcmSetActionsProperty(DeviceIntPtr dev, Atom property,
+                                 XIPropertyValuePtr prop, BOOL checkonly,
+                                 int size, Atom* handlers, unsigned int (*actions)[256])
 {
 	InputInfoPtr pInfo = (InputInfoPtr) dev->public.devicePrivate;
 	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
-	int rc;
+	int rc, i;
 
 	DBG(10, priv, "\n");
 
-	if (prop->format != 32 || prop->type != XA_ATOM)
-		return BadMatch;
+	if (prop->size != size)
+		return BadValue;
 
-	/* How this works:
-	 * prop_btnactions has a list of atoms stored. Any atom references
-	 * another property on that device that contains the actual action.
-	 * If this property changes, all action-properties are queried for
-	 * their value and their value is stored in priv->key[button].
-	 *
-	 * If the button is pressed, the actions are executed.
-	 *
-	 * Any button action property needs to be monitored by this property
-	 * handler too.
-	 */
-
-	rc = wcmCheckActionProp(dev, property, prop);
+	rc = wcmCheckActionsProperty(dev, property, prop);
 	if (rc != Success)
 		return rc;
 
-	if (!checkonly)
+	for (i = 0; i < prop->size; i++)
 	{
-		wcmUpdateActionPropHandlers(prop, priv->btn_actions);
-		wcmUpdateButtonKeyActions(dev, prop, priv->keys, ARRAY_SIZE(priv->keys));
+		int index = i;
+		Atom subproperty = ((Atom*)prop->data)[i];
+		XIPropertyValuePtr subprop;
 
-	}
-	return Success;
-}
-
-struct wheel_strip_update_t {
-	/* for CARD8 values, points to fields in struct to be updated */
-	int *up1;
-	int *dn1;
-	int *up2;
-	int *dn2;
-	int *up3;
-	int *dn3;
-
-	/* for CARD32 values, points to atom array of atoms to be
-	 * monitored.*/
-	Atom *handlers;
-	/* for CARD32 values, points to key array that keeps the actual
-	   actions.*/
-	int skeys;  /* size of first keys dimensions */
-	unsigned int (*keys)[256];
-};
-
-static int wcmSetWheelOrStripProperty(DeviceIntPtr dev, Atom property,
-				      XIPropertyValuePtr prop, BOOL checkonly,
-				      struct wheel_strip_update_t *wsup)
-{
-	int rc;
-
-	union multival {
-		CARD8 *v8;
-		CARD32 *v32;
-	} values;
-
-	if ((property == prop_strip_buttons && prop->size != 4) ||
-	    (property == prop_wheel_buttons && prop->size != 6))
-		return BadValue;
-
-	/* see wcmSetPropertyButtonActions for how this works. The wheel is
-	 * slightly different in that it allows for 8 bit properties for
-	 * pure buttons too */
-
-	values.v8 = (CARD8*)prop->data;
-
-	switch (prop->format)
-	{
-		case 8:
-			if (values.v8[0] > WCM_MAX_BUTTONS ||
-			    values.v8[1] > WCM_MAX_BUTTONS ||
-			    values.v8[2] > WCM_MAX_BUTTONS ||
-			    values.v8[3] > WCM_MAX_BUTTONS ||
-			    values.v8[4] > WCM_MAX_BUTTONS ||
-			    values.v8[5] > WCM_MAX_BUTTONS)
-				return BadValue;
-
-			if (!checkonly) {
-				*wsup->up1 = values.v8[0];
-				*wsup->dn1 = values.v8[1];
-				*wsup->up2 = values.v8[2];
-				*wsup->dn2 = values.v8[3];
-				*wsup->up3 = values.v8[4];
-				*wsup->dn3 = values.v8[5];
-			}
-			break;
-		case 32:
-			rc = wcmCheckActionProp(dev, property, prop);
-			if (rc != Success)
-				return rc;
+		if (property == prop_btnactions)
+		{ /* Driver uses physical -- not X11 -- button numbering internally */
+			if (i < 3)
+				index = i;
+			else if (i < 7)
+				continue;
+			else
+				index = i - 4;
+		}
 
+		if (subproperty == 0)
+		{ /* Interpret 'None' as meaning 'reset' */
 			if (!checkonly)
 			{
-				wcmUpdateActionPropHandlers(prop, wsup->handlers);
-				wcmUpdateButtonKeyActions(dev, prop, wsup->keys,
-						          wsup->skeys);
-			}
+				if (property == prop_btnactions)
+					wcmResetButtonAction(pInfo, index, size);
+				else if (property == prop_strip_buttons)
+					wcmResetStripAction(pInfo, index);
+				else if (property == prop_wheel_buttons)
+					wcmResetWheelAction(pInfo, index);
 
-			break;
-		default:
-			return BadMatch;
+				if (subproperty != handlers[index])
+					subproperty = handlers[index];
+			}
+		}
+		else
+		{
+			XIGetDeviceProperty(dev, subproperty, &subprop);
+			rc = wcmSetActionProperty(dev, subproperty, subprop, checkonly, &handlers[index], &actions[index]);
+			if (rc != Success)
+				return rc;
+		}
 	}
 
 	return Success;
 }
 
-
-static int wcmSetWheelProperty(DeviceIntPtr dev, Atom property,
-			       XIPropertyValuePtr prop, BOOL checkonly)
-{
-	InputInfoPtr pInfo = (InputInfoPtr) dev->public.devicePrivate;
-	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
-
-	struct wheel_strip_update_t wsup = {
-		.up1 = &priv->relup,
-		.dn1 = &priv->reldn,
-		.up2 = &priv->wheelup,
-		.dn2 = &priv->wheeldn,
-		.up3 = &priv->wheel2up,
-		.dn3 = &priv->wheel2dn,
-
-		.handlers = priv->wheel_actions,
-		.keys	  = priv->wheel_keys,
-		.skeys    = 6,
-	};
-
-	return wcmSetWheelOrStripProperty(dev, property, prop, checkonly, &wsup);
-}
-
-static int wcmSetStripProperty(DeviceIntPtr dev, Atom property,
-			       XIPropertyValuePtr prop, BOOL checkonly)
-{
-	InputInfoPtr pInfo = (InputInfoPtr) dev->public.devicePrivate;
-	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
-
-	struct wheel_strip_update_t wsup = {
-		.up1 = &priv->striplup,
-		.dn1 = &priv->stripldn,
-		.up2 = &priv->striprup,
-		.dn2 = &priv->striprdn,
-		.up3 = NULL,
-		.dn3 = NULL,
-
-		.handlers = priv->strip_actions,
-		.keys	  = priv->strip_keys,
-		.skeys    = 4,
-	};
-
-	return wcmSetWheelOrStripProperty(dev, property, prop, checkonly, &wsup);
-}
-
 /**
  * Update the rotation property for all tools on the same physical tablet as
  * pInfo.
@@ -605,6 +635,65 @@ void wcmUpdateRotationProperty(WacomDevicePtr priv)
 	}
 }
 
+static void
+wcmSetHWTouchProperty(InputInfoPtr pInfo)
+{
+	WacomDevicePtr priv = pInfo->private;
+	WacomCommonPtr common = priv->common;
+	XIPropertyValuePtr prop;
+	CARD8 prop_value;
+	int rc;
+
+	rc = XIGetDeviceProperty(pInfo->dev, prop_hardware_touch, &prop);
+	if (rc != Success || prop->format != 8 || prop->size != 1)
+	{
+		xf86Msg(X_ERROR, "%s: Failed to update hardware touch state.\n",
+			pInfo->name);
+		return;
+	}
+
+	prop_value = common->wcmHWTouchSwitchState;
+	XIChangeDeviceProperty(pInfo->dev, prop_hardware_touch, XA_INTEGER,
+			       prop->format, PropModeReplace,
+			       prop->size, &prop_value, TRUE);
+}
+
+static CARD32
+touchTimerFunc(OsTimerPtr timer, CARD32 now, pointer arg)
+{
+	InputInfoPtr pInfo = arg;
+#if !HAVE_THREADED_INPUT
+	int sigstate = xf86BlockSIGIO();
+#endif
+
+	wcmSetHWTouchProperty(pInfo);
+
+#if !HAVE_THREADED_INPUT
+	xf86UnblockSIGIO(sigstate);
+#endif
+
+	return 0;
+}
+
+/**
+ * Update HW touch property when its state is changed by touch switch
+ */
+void
+wcmUpdateHWTouchProperty(WacomDevicePtr priv, int hw_touch)
+{
+	WacomCommonPtr common = priv->common;
+
+	if (hw_touch == common->wcmHWTouchSwitchState)
+		return;
+
+	common->wcmHWTouchSwitchState = hw_touch;
+
+	/* This function is called during SIGIO/InputThread. Schedule timer
+	 * for property event delivery by the main thread. */
+	priv->touch_timer = TimerSet(priv->touch_timer, 0 /* reltime */,
+				      1, touchTimerFunc, priv->pInfo);
+}
+
 /**
  * Only allow deletion of a property if it is not being used by any of the
  * button actions.
@@ -649,8 +738,8 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 			if ((values[0] == -1) && (values[1] == -1) &&
 					(values[2] == -1) && (values[3] == -1))
 			{
-				values[0] = 0;
-				values[1] = 0;
+				values[0] = priv->minX;
+				values[1] = priv->minX;
 				values[2] = priv->maxX;
 				values[3] = priv->maxY;
 			}
@@ -737,9 +826,9 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 			wcmBindToSerial(pInfo, serial);
 		}
 	} else if (property == prop_strip_buttons)
-		return wcmSetStripProperty(dev, property, prop, checkonly);
+		return wcmSetActionsProperty(dev, property, prop, checkonly, ARRAY_SIZE(priv->strip_actions), priv->strip_actions, priv->strip_keys);
 	else if (property == prop_wheel_buttons)
-		return wcmSetWheelProperty(dev, property, prop, checkonly);
+		return wcmSetActionsProperty(dev, property, prop, checkonly, ARRAY_SIZE(priv->wheel_actions), priv->wheel_actions, priv->wheel_keys);
 	else if (property == prop_cursorprox)
 	{
 		CARD32 value;
@@ -759,6 +848,7 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 			common->wcmCursorProxoutDist = value;
 	} else if (property == prop_threshold)
 	{
+		const INT32 MAXIMUM = wcmInternalToUserPressure(pInfo, priv->maxCurve);
 		INT32 value;
 
 		if (prop->size != 1 || prop->format != 32)
@@ -767,9 +857,11 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 		value = *(INT32*)prop->data;
 
 		if (value == -1)
-			value = DEFAULT_THRESHOLD;
-		else if ((value < 1) || (value > FILTER_PRESSURE_RES))
+			value = priv->maxCurve * DEFAULT_THRESHOLD;
+		else if ((value < 1) || (value > MAXIMUM))
 			return BadValue;
+		else
+			value = wcmUserToInternalPressure(pInfo, value);
 
 		if (!checkonly)
 			common->wcmThreshold = value;
@@ -785,6 +877,19 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 
 		if (!checkonly && common->wcmTouch != values[0])
 			common->wcmTouch = values[0];
+	} else if (property == prop_hardware_touch)
+	{
+		if (common->wcmHasHWTouchSwitch)
+		{
+			/* If we get here from wcmUpdateHWTouchProperty, we know
+			 * the wcmHWTouchSwitchState has been set internally
+			 * already, so we can reply with success. */
+			if (prop->size == 1 && prop->format == 8)
+				if (((CARD8*)prop->data)[0] == common->wcmHWTouchSwitchState)
+					return Success;
+		}
+
+		return BadValue; /* read-only */
 	} else if (property == prop_gesture)
 	{
 		CARD8 *values = (CARD8*)prop->data;
@@ -850,12 +955,31 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 #endif
 	} else if (property == prop_btnactions)
 	{
-		int nbuttons = min(max(priv->nbuttons + 4, 7), WCM_MAX_BUTTONS);
-		if (prop->size != nbuttons)
+		int nbuttons = priv->nbuttons < 4 ? priv->nbuttons : priv->nbuttons + 4;
+		return wcmSetActionsProperty(dev, property, prop, checkonly, nbuttons, priv->btn_actions, priv->keys);
+	} else if (property == prop_pressure_recal)
+	{
+		CARD8 *values = (CARD8*)prop->data;
+
+		if (prop->size != 1 || prop->format != 8)
+			return BadValue;
+
+		if ((values[0] != 0) && (values[0] != 1))
+			return BadValue;
+
+		if (!IsStylus(priv) && !IsEraser(priv))
 			return BadMatch;
-		wcmSetPropertyButtonActions(dev, property, prop, checkonly);
+
+		if (!checkonly)
+			common->wcmPressureRecalibration = values[0];
 	} else
-		wcmSetActionProperties(dev, property, prop, checkonly);
+	{
+		Atom *handler = NULL;
+		unsigned int (*action)[256] = NULL;
+		if (wcmFindActionHandler(priv, property, &handler, &action))
+			return wcmSetActionProperty(dev, property, prop, checkonly, handler, action);
+		/* backwards-compatible behavior silently ignores the not-found case */
+	}
 
 	return Success;
 }
@@ -873,39 +997,72 @@ int wcmGetProperty (DeviceIntPtr dev, Atom property)
 		uint32_t values[5];
 
 		values[0] = common->tablet_id;
-		values[1] = priv->old_serial;
-		values[2] = priv->old_device_id;
+		values[1] = priv->oldState.serial_num;
+		values[2] = priv->oldState.device_id;
 		values[3] = priv->cur_serial;
 		values[4] = priv->cur_device_id;
 
-		DBG(10, priv, "Update to serial: %d\n", priv->old_serial);
+		DBG(10, priv, "Update to serial: %d\n", priv->oldState.serial_num);
 
 		return XIChangeDeviceProperty(dev, property, XA_INTEGER, 32,
 					      PropModeReplace, 5,
 					      values, FALSE);
 	}
+	else if (property == prop_btnactions)
+	{
+		/* Convert the physical button representation used internally
+		 * to the X11 button representation we've historically used.
+		 * To do this, we need to skip X11 buttons 4-7 which would be
+		 * used by a scroll wheel rather than an actual button.
+		 */
+		int nbuttons = priv->nbuttons < 4 ? priv->nbuttons : priv->nbuttons + 4;
+		Atom x11_btn_actions[nbuttons];
+		int i;
+
+		for (i = 0; i < nbuttons; i++)
+		{
+			if (i < 3)
+				x11_btn_actions[i] = priv->btn_actions[i];
+			else if (i < 7)
+				x11_btn_actions[i] = 0;
+			else
+				x11_btn_actions[i] = priv->btn_actions[i-4];
+		}
+
+		return XIChangeDeviceProperty(dev, property, XA_ATOM, 32,
+		                              PropModeReplace, nbuttons,
+		                              x11_btn_actions, FALSE);
+	}
+	else if (property == prop_strip_buttons)
+	{
+		return XIChangeDeviceProperty(dev, property, XA_ATOM, 32,
+					      PropModeReplace, ARRAY_SIZE(priv->strip_actions),
+					      priv->strip_actions, FALSE);
+	}
+	else if (property == prop_wheel_buttons)
+	{
+		return XIChangeDeviceProperty(dev, property, XA_ATOM, 32,
+		                              PropModeReplace, ARRAY_SIZE(priv->wheel_actions),
+		                              priv->wheel_actions, FALSE);
+	}
 
 	return Success;
 }
 
-static CARD32
-serialTimerFunc(OsTimerPtr timer, CARD32 now, pointer arg)
+static void
+wcmSetSerialProperty(InputInfoPtr pInfo)
 {
-	InputInfoPtr pInfo = arg;
 	WacomDevicePtr priv = pInfo->private;
 	XIPropertyValuePtr prop;
 	CARD32 prop_value[5];
-	int sigstate;
 	int rc;
 
-	sigstate = xf86BlockSIGIO();
-
 	rc = XIGetDeviceProperty(pInfo->dev, prop_serials, &prop);
 	if (rc != Success || prop->format != 32 || prop->size != 5)
 	{
 		xf86Msg(X_ERROR, "%s: Failed to update serial number.\n",
 			pInfo->name);
-		return 0;
+		return;
 	}
 
 	memcpy(prop_value, prop->data, sizeof(prop_value));
@@ -915,8 +1072,22 @@ serialTimerFunc(OsTimerPtr timer, CARD32 now, pointer arg)
 	XIChangeDeviceProperty(pInfo->dev, prop_serials, XA_INTEGER,
 			       prop->format, PropModeReplace,
 			       prop->size, prop_value, TRUE);
+}
 
+static CARD32
+serialTimerFunc(OsTimerPtr timer, CARD32 now, pointer arg)
+{
+	InputInfoPtr pInfo = arg;
+
+#if !HAVE_THREADED_INPUT
+	int sigstate = xf86BlockSIGIO();
+#endif
+
+	wcmSetSerialProperty(pInfo);
+
+#if !HAVE_THREADED_INPUT
 	xf86UnblockSIGIO(sigstate);
+#endif
 
 	return 0;
 }
@@ -926,14 +1097,14 @@ wcmUpdateSerial(InputInfoPtr pInfo, unsigned int serial, int id)
 {
 	WacomDevicePtr priv = pInfo->private;
 
-	if (priv->cur_serial == serial)
+	if (priv->cur_serial == serial && priv->cur_device_id == id)
 		return;
 
 	priv->cur_serial = serial;
 	priv->cur_device_id = id;
 
-	/* This function is called during SIGIO. Schedule timer for property
-	 * event delivery outside of signal handler. */
+	/* This function is called during SIGIO/InputThread. Schedule timer
+	 * for property event delivery by the main thread. */
 	priv->serial_timer = TimerSet(priv->serial_timer, 0 /* reltime */,
 				      1, serialTimerFunc, pInfo);
 }
