@@ -200,7 +200,7 @@ static int wcmInitAxes(DeviceIntPtr pWcm)
 	if (!IsPad(priv))
 	{
 		label = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_PRESSURE);
-		max = FILTER_PRESSURE_RES;
+		max = priv->maxCurve;
 	}
 
 	wcmInitAxis(pInfo->dev, index, label, min, max, res, min_res, max_res, mode);
@@ -282,8 +282,8 @@ static int wcmInitAxes(DeviceIntPtr pWcm)
 	{
 		/* Touch ring */
 		label = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_WHEEL);
-		min = MIN_PAD_RING;
-		max = MAX_PAD_RING;
+		min = common->wcmMinRing;
+		max = common->wcmMaxRing;
 	}
 
 	wcmInitAxis(pInfo->dev, index, label, min, max, res, min_res, max_res, mode);
@@ -298,8 +298,8 @@ static int wcmInitAxes(DeviceIntPtr pWcm)
 		mode = Absolute;
 		min_res = max_res = res = 1;
 
-		min = MIN_PAD_RING;
-		max = MAX_PAD_RING;
+		min = common->wcmMinRing;
+		max = common->wcmMaxRing;
 
 		wcmInitAxis(pInfo->dev, index, label, min, max, res, min_res, max_res, mode);
 	}
@@ -341,12 +341,6 @@ static int wcmDevInit(DeviceIntPtr pWcm)
 	nbkeys = nbbuttons;         /* Same number of keys since any button may be 
 	                             * configured as an either mouse button or key */
 
-	if (!nbbuttons)
-		nbbuttons = nbkeys = 1;	    /* Xserver 1.5 or later crashes when 
-			            	     * nbbuttons = 0 while sending a beep 
-			             	     * This is only a workaround. 
-				     	     */
-
 	DBG(10, priv,
 		"(%s) %d buttons, %d keys, %d axes\n",
 		pInfo->type_name,
@@ -753,9 +747,9 @@ static void wcmDevClose(InputInfoPtr pInfo)
 
 	if (pInfo->fd >= 0)
 	{
-		pInfo->fd = -1;
 		if (!--common->fd_refs)
 			wcmClose(pInfo);
+		pInfo->fd = -1;
 	}
 }
 
@@ -829,9 +823,7 @@ static void wcmUnlinkTouchAndPen(InputInfoPtr pInfo)
 static int wcmDevProc(DeviceIntPtr pWcm, int what)
 {
 	InputInfoPtr pInfo = (InputInfoPtr)pWcm->public.devicePrivate;
-#ifdef DEBUG
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
-#endif
 	Status rc = !Success;
 
 	DBG(2, priv, "BEGIN dev=%p priv=%p "
@@ -860,7 +852,9 @@ static int wcmDevProc(DeviceIntPtr pWcm, int what)
 			break;
 
 		case DEVICE_OFF:
-		case DEVICE_CLOSE:
+			TimerCancel(priv->tap_timer);
+			TimerCancel(priv->serial_timer);
+			TimerCancel(priv->touch_timer);
 			wcmDisableTool(pWcm);
 			wcmUnlinkTouchAndPen(pInfo);
 			if (pInfo->fd >= 0)
@@ -870,6 +864,8 @@ static int wcmDevProc(DeviceIntPtr pWcm, int what)
 			}
 			pWcm->public.on = FALSE;
 			break;
+		case DEVICE_CLOSE:
+			break;
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) * 100 + GET_ABI_MINOR(ABI_XINPUT_VERSION) >= 1901
 		case DEVICE_ABORT:
 			break;
