@@ -1,6 +1,6 @@
 /*
  * Copyright 1995-2002 by Frederic Lepied, France. <Lepied@XFree86.org>
- * Copyright 2002-2010 by Ping Cheng, Wacom. <pingc@wacom.com>
+ * Copyright 2002-2013 by Ping Cheng, Wacom. <pingc@wacom.com>
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -43,8 +43,10 @@
 #define MAX_ROTATION_RANGE 1800 /* the maximum range of the marker pen rotation */
 #define MAX_ABS_WHEEL 1023      /* the maximum value of absolute wheel */
 
-#define MIN_PAD_RING 0		/* I4 absolute scroll ring min value */
-#define MAX_PAD_RING 71		/* I4 absolute scroll ring max value */
+#define TILT_RES (180/M_PI)	/* Reported tilt resolution in points/radian
+				   (1/degree) */
+#define TILT_MIN -64		/* Minimum reported tilt value */
+#define TILT_MAX 63		/* Maximum reported tilt value */
 
 /* I4 cursor tool has a rotation offset of 175 degrees */
 #define INTUOS4_CURSOR_ROTATION_OFFSET 175
@@ -91,10 +93,6 @@
 #define ABS_MT_SLOT 0x2f
 #endif
 
-/* defines to discriminate second side button and the eraser */
-#define ERASER_PROX     4
-#define OTHER_PROX      1
-
 /******************************************************************************
  * Forward Declarations
  *****************************************************************************/
@@ -107,7 +105,6 @@ typedef struct _WacomCommonRec WacomCommonRec, *WacomCommonPtr;
 typedef struct _WacomFilterState WacomFilterState, *WacomFilterStatePtr;
 typedef struct _WacomDeviceClass WacomDeviceClass, *WacomDeviceClassPtr;
 typedef struct _WacomTool WacomTool, *WacomToolPtr;
-typedef struct _WacomToolArea WacomToolArea, *WacomToolAreaPtr;
 
 /******************************************************************************
  * WacomModel - model-specific device capabilities
@@ -165,6 +162,7 @@ struct _WacomModel
 							  button handling,
 							  always an LCD) */
 #define WCM_PENTOUCH		0x00000400 /* Tablet supports pen and touch */
+#define WCM_DUALRING		0x00000800 /* Tablet has two touch rings */
 #define TabletHasFeature(common, feature) MaskIsSet((common)->tablet_type, (feature))
 #define TabletSetFeature(common, feature) MaskSet((common)->tablet_type, (feature))
 
@@ -178,29 +176,64 @@ struct _WacomModel
 #define IsEraser(priv) (DEVICE_ID((priv)->flags) == ERASER_ID)
 #define IsPad(priv)    (DEVICE_ID((priv)->flags) == PAD_ID)
 #define IsPen(priv)    (IsStylus(priv) || IsEraser(priv))
+#define IsTablet(priv) (IsPen(priv) || IsCursor(priv))
 
 #define IsUSBDevice(common) ((common)->wcmDevCls == &gWacomUSBDevice)
 
-#define FILTER_PRESSURE_RES	2048	/* maximum points in pressure curve */
+#define FILTER_PRESSURE_RES	65536	/* maximum points in pressure curve */
+/* Tested result for setting the pressure threshold to a reasonable value */
+#define THRESHOLD_TOLERANCE (0.008f)
+#define DEFAULT_THRESHOLD (0.013f)
+
 #define WCM_MAX_BUTTONS		32	/* maximum number of tablet buttons */
-#define WCM_MAX_MOUSE_BUTTONS	16	/* maximum number of buttons-on-pointer
-                                         * (which are treated as mouse buttons,
-                                         * not as keys like tablet menu buttons). 
-					 * For backward compability support,
-					 * tablet buttons besides the strips are
-					 * treated as buttons */
+#define WCM_MAX_X11BUTTON	127	/* maximum button number X11 can handle */
 
-/* get/set/property */
-typedef struct _PROPINFO PROPINFO;
+#define AXIS_INVERT  0x01               /* Flag describing an axis which increases "downward" */
+#define AXIS_BITWISE 0x02               /* Flag describing an axis which changes bitwise */
 
-struct _PROPINFO
+/* Indicies into the wheel/strip default/keys/actions arrays */
+#define WHEEL_REL_UP      0
+#define WHEEL_REL_DN      1
+#define WHEEL_ABS_UP      2
+#define WHEEL_ABS_DN      3
+#define WHEEL2_ABS_UP     4
+#define WHEEL2_ABS_DN     5
+#define STRIP_LEFT_UP     0
+#define STRIP_LEFT_DN     1
+#define STRIP_RIGHT_UP    2
+#define STRIP_RIGHT_DN    3
+
+/******************************************************************************
+ * WacomDeviceState
+ *****************************************************************************/
+struct _WacomDeviceState
 {
-	Atom wcmProp;
-	char paramName[32];
-	int nParamID;
-	int nFormat;
-	int nSize;
-	int nDefault;
+	InputInfoPtr pInfo;
+	int device_id;		/* tool id reported from the physical device */
+	int device_type;
+	unsigned int serial_num;
+	int x;
+	int y;
+	int buttons;
+	int pressure;
+	int tiltx;
+	int tilty;
+	int stripx;
+	int stripy;
+	int rotation;
+	int abswheel;
+	int abswheel2;
+	int relwheel;
+	int distance;
+	int throttle;
+	int proximity;
+	int sample;	/* wraps every 24 days */
+	int time;
+};
+
+static const struct _WacomDeviceState OUTPROX_STATE = {
+  .abswheel = INT_MAX,
+  .abswheel2 = INT_MAX
 };
 
 struct _WacomDeviceRec
@@ -218,37 +251,30 @@ struct _WacomDeviceRec
 	int bottomY;		/* Y bottom in device coordinates */
 	int resolX;             /* X resolution */
 	int resolY;             /* Y resolution */
-	int maxX;	        /* tool physical maxX in device coordinates*/
-	int maxY;	        /* tool physical maxY in device coordinates*/
-	double factorX;		/* X factor */
-	double factorY;		/* Y factor */
+	int minX;	        /* tool physical minX in device coordinates */
+	int minY;	        /* tool physical minY in device coordinates */
+	int maxX;	        /* tool physical maxX in device coordinates */
+	int maxY;	        /* tool physical maxY in device coordinates */
 	unsigned int serial;	/* device serial number this device takes (if 0, any serial is ok) */
 	unsigned int cur_serial; /* current serial in prox */
-	int maxWidth;		/* max active screen width in screen coords */
-	int maxHeight;		/* max active screen height in screen coords */
-	int leftPadding;	/* left padding for virtual tablet in device coordinates*/
-	int topPadding;		/* top padding for virtual tablet in device coordinates*/
-	/*  map zero based internal buttons to one based X buttons */
-	int button[WCM_MAX_BUTTONS];
-	/* map one based X buttons to keystrokes */
-	unsigned keys[WCM_MAX_BUTTONS+1][256];
-	int relup;
-	int reldn;
-	int wheelup;
-	int wheeldn;
-	/* keystrokes assigned to wheel events (default is the buttons above).
-	 * Order is relup, reldwn, wheelup, wheeldn. Like 'keys', this array
-	 * is one-indexed */
-	unsigned wheel_keys[4+1][256];
+	int cur_device_id;	/* current device ID in prox */
+
+	/* button mapping information
+	 *
+	 * 'button' variables are indexed by physical button number (0..nbuttons)
+	 * 'strip' variables are indexed by STRIP_* defines
+	 * 'wheel' variables are indexed by WHEEL_* defines
+	 */
+	int button_default[WCM_MAX_BUTTONS]; /* Default mappings set by ourselves (possibly overridden by xorg.conf) */
+	int strip_default[4];
+	int wheel_default[6];
+	unsigned keys[WCM_MAX_BUTTONS][256]; /* Action codes to perform when the associated event occurs */
+	unsigned strip_keys[4][256];
+	unsigned wheel_keys[6][256];
+	Atom btn_actions[WCM_MAX_BUTTONS];   /* Action references so we can update the action codes when a client makes a change */
+	Atom strip_actions[4];
+	Atom wheel_actions[6];
 
-	int striplup;
-	int stripldn;
-	int striprup;
-	int striprdn;
-	/* keystrokes assigned to strip events (default is the buttons above).
-	 * Order is striplup, stripldn, striprup, striprdn. Like 'keys', this
-	 * array is one-indexed */
-	unsigned strip_keys[4+1][256];
 	int nbuttons;           /* number of buttons for this subdevice */
 	int naxes;              /* number of axes */
 				/* FIXME: always 6, and the code relies on that... */
@@ -256,79 +282,28 @@ struct _WacomDeviceRec
 	WacomCommonPtr common;  /* common info pointer */
 
 	/* state fields in device coordinates */
-	int currentX;           /* current X position */
-	int currentY;           /* current Y position */
-	int currentSX;          /* current screen X position in screen coords */
-	int currentSY;          /* current screen Y position in screen coords */
-	int oldX;               /* previous X position */
-	int oldY;               /* previous Y position */
-	int oldZ;               /* previous pressure */
-	int oldTiltX;           /* previous tilt in x direction */
-	int oldTiltY;           /* previous tilt in y direction */    
-	int oldWheel;           /* previous wheel value */    
-	int oldRot;             /* previous rotation value */
-	int oldStripX;          /* previous left strip value */
-	int oldStripY;          /* previous right strip value */
-	int oldThrottle;        /* previous throttle value */
-	int oldButtons;         /* previous buttons state */
-	int oldProximity;       /* previous proximity */
+	struct _WacomDeviceState oldState; /* previous state information */
 	int oldCursorHwProx;	/* previous cursor hardware proximity */
-	int old_device_id;	/* last in prox device id */
-	unsigned int old_serial;/* last in prox tool serial number */
-	int devReverseCount;	/* Relative ReverseConvert called twice each movement*/
 
-	/* JEJ - throttle */
-	int throttleStart;      /* time in ticks for last wheel movement */
-	int throttleLimit;      /* time in ticks for next wheel movement */
-	int throttleValue;      /* current throttle value */
-
-	/* JEJ - filters */
-	int pPressCurve[FILTER_PRESSURE_RES + 1]; /* pressure curve */
+	int maxCurve;		/* maximum pressure curve value */
+	int *pPressCurve;       /* pressure curve */
 	int nPressCtrl[4];      /* control points for curve */
 	int minPressure;	/* the minimum pressure a pen may hold */
-
+	int oldMinPressure;     /* to record the last minPressure before going out of proximity */
+	unsigned int eventCnt;  /* count number of events while in proximity */
+	int maxRawPressure;     /* maximum 'raw' pressure seen until first button event */
 	WacomToolPtr tool;         /* The common tool-structure for this device */
 
 	int isParent;		/* set to 1 if the device is not auto-hotplugged */
 
-	/* property handlers to listen to for action properties */
-	Atom btn_actions[WCM_MAX_BUTTONS];
-	Atom wheel_actions[4];
-	Atom strip_actions[4];
-
 	OsTimerPtr serial_timer; /* timer used for serial number property update */
+	OsTimerPtr tap_timer;   /* timer used for tap timing */
+	OsTimerPtr touch_timer; /* timer used for touch switch property update */
 };
 
-/******************************************************************************
- * WacomDeviceState
- *****************************************************************************/
-
 #define MAX_SAMPLES	20
 #define DEFAULT_SAMPLES 4
 
-struct _WacomDeviceState
-{
-	InputInfoPtr pInfo;
-	int device_id;		/* tool id reported from the physical device */
-	int device_type;
-	unsigned int serial_num;
-	int x;
-	int y;
-	int buttons;
-	int pressure;
-	int tiltx;
-	int tilty;
-	int stripx;
-	int stripy;
-	int rotation;
-	int abswheel;
-	int relwheel;
-	int distance;
-	int throttle;
-	int proximity;
-	int sample;	/* wraps every 24 days */
-};
-
 struct _WacomFilterState
 {
         int npoints;
@@ -346,6 +321,7 @@ struct _WacomChannel
 	 * the work stage and the valid state. */
 
 	WacomDeviceState work;                         /* next state */
+	Bool dirty;
 
 	/* the following union contains the current known state of the
 	 * device channel, as well as the previous MAX_SAMPLES states
@@ -368,7 +344,7 @@ struct _WacomDeviceClass
 {
 	Bool (*Detect)(InputInfoPtr pInfo); /* detect device */
 	Bool (*ParseOptions)(InputInfoPtr pInfo); /* parse class-specific options */
-	Bool (*Init)(InputInfoPtr pInfo, char* id, float *version);   /* initialize device */
+	Bool (*Init)(InputInfoPtr pInfo, char* id, size_t id_len, float *version);   /* initialize device */
 	int  (*ProbeKeys)(InputInfoPtr pInfo); /* set the bits for the keys supported */
 };
 
@@ -379,22 +355,19 @@ extern WacomDeviceClass gWacomISDV4Device;
  * WacomCommonRec
  *****************************************************************************/
 
-#define TILT_REQUEST_FLAG       1
 #define TILT_ENABLED_FLAG       2
 
-#define MAX_CHANNELS 3
+#define MAX_FINGERS 16
+#define MAX_CHANNELS (MAX_FINGERS+2) /* one channel for stylus/mouse. The other one for pad */
 #define PAD_CHANNEL (MAX_CHANNELS-1)
-#define MAX_FINGERS  2
 
 typedef struct {
 	int wcmZoomDistance;	       /* minimum distance for a zoom touch gesture */
-	int wcmZoomDistanceDefault;    /* default minimum distance for a zoom touch gesture */
 	int wcmScrollDistance;	       /* minimum motion before sending a scroll gesture */
 	int wcmScrollDirection;	       /* store the vertical or horizontal bit in use */
-	int wcmScrollDistanceDefault;  /* default minimum motion before sending a scroll gesture */
+	int wcmMaxScrollFingerSpread; /* maximum distance between fingers for scroll gesture */
 	int wcmGestureUsed;	       /* retain used gesture count within one in-prox event */
 	int wcmTapTime;	   	       /* minimum time between taps for a right click */
-	int wcmTapTimeDefault;         /* default minimum time between taps for a right click */
 } WacomGesturesParameters;
 
 enum WacomProtocol {
@@ -403,10 +376,16 @@ enum WacomProtocol {
 	WCM_PROTOCOL_5
 };
 
+struct _WacomDriverRec
+{
+	WacomDevicePtr active;     /* Arbitrate motion through this pointer */
+};
+extern struct _WacomDriverRec WACOM_DRIVER; // Defined in wcmCommon.c
+
 struct _WacomCommonRec 
 {
 	/* Do not move device_path, same offset as priv->name. Used by DBG macro */
-	char* device_path;           /* device file name */
+	char* device_path;          /* device file name */
 	dev_t min_maj;               /* minor/major number */
 	unsigned char wcmFlags;     /* various flags (handle tilt) */
 	int debugLevel;
@@ -418,9 +397,13 @@ struct _WacomCommonRec
 	unsigned long wcmKeys[NBITS(KEY_MAX)]; /* supported tool types for the device */
 	WacomDevicePtr wcmTouchDevice; /* The pointer for pen to access the
 					  touch tool of the same device id */
-	Bool wcmPenInProx;      /* Keep pen in-prox state for touch tool */
+
+	Bool wcmHasHWTouchSwitch;    /* Tablet has a touch on/off switch */
+	int wcmHWTouchSwitchState;   /* touch event disable/enabled by hardware switch */
 
 	/* These values are in tablet coordinates */
+	int wcmMinX;                 /* tablet min X value */
+	int wcmMinY;                 /* tablet min Y value */
 	int wcmMaxX;                 /* tablet max X value */
 	int wcmMaxY;                 /* tablet max Y value */
 	int wcmMaxZ;                 /* tablet max Z value */
@@ -433,15 +416,25 @@ struct _WacomCommonRec
 	                             /* tablet Z resolution is equivalent
 	                              * to wcmMaxZ which is equal to 100% pressure */
 	int wcmMaxDist;              /* tablet max distance value */
-	int wcmMaxtiltX;	     /* styli max tilt in X directory */ 
-	int wcmMaxtiltY;	     /* styli max tilt in Y directory */ 
+	int wcmMaxContacts;          /* MT device max number of contacts */
+
+	/*
+	 * TODO Remove wcmTiltOff*, once the kernel drivers reporting
+	 * 	non-zero-centered tilt values are no longer in use.
+	 */
+	int wcmTiltOffX;	     /* styli tilt offset in X direction */
+	int wcmTiltOffY;	     /* styli tilt offset in Y direction */
+	double wcmTiltFactX;	     /* styli tilt factor in X direction */
+	double wcmTiltFactY;	     /* styli tilt factor in Y direction */
+	int wcmTiltMinX;	     /* styli min reported tilt in X direction */
+	int wcmTiltMinY;	     /* styli min reported tilt in Y direction */
+	int wcmTiltMaxX;	     /* styli max reported tilt in X direction */
+	int wcmTiltMaxY;	     /* styli max reported tilt in Y direction */
 
 	int wcmMaxStripX;            /* Maximum fingerstrip X */
 	int wcmMaxStripY;            /* Maximum fingerstrip Y */
-
-	int nbuttons;                /* total number of buttons */
-	int padkey_code[WCM_MAX_BUTTONS];/* hardware codes for buttons */
-	int npadkeys;                /* number of pad keys in the above array */
+	int wcmMinRing;              /* Minimum touchring value */
+	int wcmMaxRing;              /* Maximum touchring value */
 
 	WacomDevicePtr wcmDevices;   /* list of devices sharing same port */
 	int wcmPktLength;            /* length of a packet */
@@ -466,6 +459,8 @@ struct _WacomCommonRec
 	int wcmCursorProxoutDistDefault; /* Default max mouse distance for proxy-out */
 	int wcmSuppress;        	 /* transmit position on delta > supress */
 	int wcmRawSample;	     /* Number of raw data used to filter an event */
+	int wcmPressureRecalibration; /* Determine if pressure recalibration of
+					 worn pens should be performed */
 
 	int bufpos;                        /* position with buffer */
 	unsigned char buffer[BUFFER_SIZE]; /* data read from device */
@@ -477,6 +472,10 @@ struct _WacomCommonRec
 
 	/* DO NOT TOUCH THIS. use wcmRefCommon() instead */
 	int refcnt;			/* number of devices sharing this struct */
+
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 16
+	ValuatorMask *touch_mask;
+#endif
 };
 
 #define HANDLE_TILT(comm) ((comm)->wcmFlags & TILT_ENABLED_FLAG)
